import logging
import os
import re
from itertools import groupby
from typing import List, Optional, Tuple

from uaclient import api, apt, event_logger, exceptions, messages, system, util
from uaclient.clouds.identity import NoCloudTypeReason, get_cloud_type
from uaclient.entitlements import repo
from uaclient.entitlements.base import EntitlementWithMessage
from uaclient.entitlements.entitlement_status import ApplicationStatus
from uaclient.files import notices
from uaclient.files.notices import Notice
from uaclient.files.state_files import (
    ServicesOnceEnabledData,
    services_once_enabled_file,
)
from uaclient.types import (  # noqa: F401
    MessagingOperations,
    MessagingOperationsDict,
    StaticAffordance,
)

event = event_logger.get_event_logger()
LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))

CONDITIONAL_PACKAGES_EVERYWHERE = [
    "strongswan",
    "strongswan-hmac",
    "openssh-client",
    "openssh-server",
    "shim-signed",
]
CONDITIONAL_PACKAGES_OPENSSH_HMAC = [
    "openssh-client-hmac",
    "openssh-server-hmac",
]
FIPS_CONDITIONAL_PACKAGES = {
    "xenial": CONDITIONAL_PACKAGES_EVERYWHERE
    + CONDITIONAL_PACKAGES_OPENSSH_HMAC,
    "bionic": CONDITIONAL_PACKAGES_EVERYWHERE
    + CONDITIONAL_PACKAGES_OPENSSH_HMAC,
    "focal": CONDITIONAL_PACKAGES_EVERYWHERE,
}


# In containers, we don't install the ubuntu-fips
# metapackage, but we do want to auto-upgrade any
# fips related packages that are already installed.
# These lists need to be kept up to date with the
# Depends of ubuntu-fips.
# Note that these lists only include those packages
# that are relevant for a container to upgrade
# after enabling fips or fips-updates.
UBUNTU_FIPS_METAPACKAGE_DEPENDS_XENIAL = [
    "openssl",
    "libssl1.0.0",
    "libssl1.0.0-hmac",
]
UBUNTU_FIPS_METAPACKAGE_DEPENDS_BIONIC = [
    "openssl",
    "libssl1.1",
    "libssl1.1-hmac",
    "libgcrypt20",
    "libgcrypt20-hmac",
]
UBUNTU_FIPS_METAPACKAGE_DEPENDS_FOCAL = [
    "openssl",
    "libssl1.1",
    "libssl1.1-hmac",
    "libgcrypt20",
    "libgcrypt20-hmac",
]
FIPS_CONTAINER_CONDITIONAL_PACKAGES = {
    "xenial": CONDITIONAL_PACKAGES_EVERYWHERE
    + CONDITIONAL_PACKAGES_OPENSSH_HMAC
    + UBUNTU_FIPS_METAPACKAGE_DEPENDS_XENIAL,
    "bionic": CONDITIONAL_PACKAGES_EVERYWHERE
    + CONDITIONAL_PACKAGES_OPENSSH_HMAC
    + UBUNTU_FIPS_METAPACKAGE_DEPENDS_BIONIC,
    "focal": CONDITIONAL_PACKAGES_EVERYWHERE
    + UBUNTU_FIPS_METAPACKAGE_DEPENDS_FOCAL,
}


class FIPSCommonEntitlement(repo.RepoEntitlement):
    repo_pin_priority = 1001
    repo_key_file = "ubuntu-pro-fips.gpg"  # Same for fips & fips-updates
    FIPS_PROC_FILE = "/proc/sys/crypto/fips_enabled"
    pre_enable_msg = messages.PROMPT_FIPS_PRE_ENABLE
    supports_access_only = True

    # RELEASE_BLOCKER GH: #104, don't prompt for conf differences in FIPS
    # Review this fix to see if we want more general functionality for all
    # services. And security/CPC signoff on expected conf behavior.
    apt_noninteractive = True

    help_doc_url = messages.urls.FIPS_HOME_PAGE

    fips_pro_package_holds = [
        "fips-initramfs",
        "fips-initramfs-generic",
        "libgcrypt20",
        "libgcrypt20-hmac",
        "libgmp10",
        "libgnutls30",
        "libhogweed6",
        "libnettle8",
        "libssl1.0.0",
        "libssl1.0.0-hmac",
        "libssl1.0.0",
        "libssl1.0.0-hmac",
        "libssl1.1",
        "libssl1.1-hmac",
        "libssl3",
        "linux-fips",
        "openssh-client",
        "openssh-client-hmac",
        "openssh-server",
        "openssh-server-hmac",
        "openssl",
        "openssl-fips-module-3",
        "shim-signed",
        "strongswan",
        "strongswan-hmac",
        "ubuntu-fips",
        "ubuntu-aws-fips",
        "ubuntu-azure-fips",
        "ubuntu-gcp-fips",
    ]

    @property
    def messaging(self) -> MessagingOperationsDict:
        pre_enable = None  # type: Optional[MessagingOperations]
        post_enable = None  # type: Optional[MessagingOperations]

        if not self.access_only:
            if system.is_container():
                pre_enable_prompt = (
                    messages.PROMPT_FIPS_CONTAINER_PRE_ENABLE.format(
                        title=self.title
                    )
                )
                if not self.auto_upgrade_all_on_enable():
                    post_enable = [messages.FIPS_RUN_APT_UPGRADE]
            else:
                pre_enable_prompt = self.pre_enable_msg
            pre_enable = [
                (
                    util.prompt_for_confirmation,
                    {"msg": pre_enable_prompt},
                )
            ]

        pre_disable = None  # type: Optional[MessagingOperations]
        if not self.purge and not self.access_only:
            pre_disable = [
                (
                    util.prompt_for_confirmation,
                    {
                        "msg": messages.PROMPT_FIPS_PRE_DISABLE.format(
                            title=self.title
                        ),
                    },
                )
            ]

        pre_install = (
            [
                (
                    self.prompt_if_kernel_downgrade,
                    {},
                )
            ]
            if not self.access_only
            else None
        )  # type: Optional[MessagingOperations]

        messaging = {
            "pre_enable": pre_enable,
            "pre_install": pre_install,
            "post_enable": post_enable,
            "pre_disable": pre_disable,
        }  # type: MessagingOperationsDict

        if len(self.packages) == 1:
            # that is the kernel "ubuntu-fips" or "ubuntu-${cloud}-fips"
            ubuntu_fips_package_name = self.packages[0]
            ubuntu_fips_package_flavor_match = re.match(
                "ubuntu-([a-z]+)-fips", ubuntu_fips_package_name
            )
            if ubuntu_fips_package_flavor_match:
                ubuntu_fips_package_flavor = (
                    ubuntu_fips_package_flavor_match.group(1)
                )
            else:
                ubuntu_fips_package_flavor = "generic"
            current_flavor = system.get_kernel_info().flavor
            if ubuntu_fips_package_flavor != current_flavor:
                pre_enable = messaging.get("pre_enable") or []
                msg = messages.KERNEL_FLAVOR_CHANGE_WARNING_PROMPT.format(
                    variant=ubuntu_fips_package_flavor,
                    service=self.name,
                    base_flavor=ubuntu_fips_package_flavor,
                    current_flavor=current_flavor or "unknown",
                )
                pre_enable.append(
                    (
                        util.prompt_for_confirmation,
                        {"msg": msg},
                    )
                )
                messaging["pre_enable"] = pre_enable

        return messaging

    @property
    def conditional_packages(self):
        """
        Dictionary of conditional packages to be installed when
        enabling FIPS services. For example, if we are enabling
        FIPS services in a machine that has openssh-client installed,
        we will perform two actions:

        1. Upgrade the package to the FIPS version
        2. Install the corresponding hmac version of that package
           when available.
        """
        series = system.get_release_info().series

        if system.is_container():
            return FIPS_CONTAINER_CONDITIONAL_PACKAGES.get(series, [])

        return FIPS_CONDITIONAL_PACKAGES.get(series, [])

    def prompt_if_kernel_downgrade(self, *, assume_yes: bool) -> bool:
        """Check if installing a FIPS kernel will downgrade the kernel
        and prompt for confirmation if it will.
        """
        # Prior to installing packages, check if the kernel is being downgraded
        # and if so verify that the user wants to continue
        our_full_kernel_str = (
            system.get_kernel_info().proc_version_signature_version
        )
        if our_full_kernel_str is None:
            LOG.warning("Cannot gather kernel information")
            return False
        our_m = re.search(
            r"(?P<kernel_version>\d+\.\d+\.\d+)", our_full_kernel_str
        )
        # This relies on the fact that fips has a single metapackage in
        # additionalPackages. If more packages are added to additionalPackages
        # for some reason, we fall back to "ubuntu-fips".
        fips_package = "ubuntu-fips"
        if len(self.packages) == 1:
            fips_package = self.packages[0]

        # We are interested in the actual kernel version
        fips_kernel_version_str = apt.get_pkg_candidate_version(
            fips_package.replace("ubuntu", "linux")
        )
        if our_m is not None and fips_kernel_version_str is not None:
            our_kernel_version_str = our_m.group("kernel_version")
            LOG.debug(
                "Kernel information: cur='%s' and fips='%s'",
                our_full_kernel_str,
                fips_kernel_version_str,
            )
            if (
                apt.version_compare(
                    fips_kernel_version_str, our_kernel_version_str
                )
                < 0
            ):
                event.info(
                    messages.KERNEL_DOWNGRADE_WARNING.format(
                        current_version=our_kernel_version_str,
                        new_version=fips_kernel_version_str,
                    )
                )
                return util.prompt_for_confirmation(
                    msg=messages.PROMPT_YES_NO, assume_yes=assume_yes
                )
        else:
            LOG.warning(
                "Cannot gather kernel information for '%s' and '%s'",
                our_full_kernel_str,
                fips_kernel_version_str,
            )
        return True

    def hardcoded_install_conditional_packages(
        self, progress: api.ProgressWrapper
    ):
        # Any conditional packages should still be installed, but if
        # they fail to install we should not block the enable operation.
        desired_packages = []  # type: List[str]
        installed_packages = apt.get_installed_packages_names()
        pkg_groups = groupby(
            sorted(self.conditional_packages),
            key=lambda pkg_name: pkg_name.replace("-hmac", ""),
        )

        for pkg_name, pkg_list in pkg_groups:
            if pkg_name in installed_packages:
                desired_packages += pkg_list

        for pkg in desired_packages:
            try:
                apt.run_apt_install_command(
                    packages=[pkg],
                    override_env_vars={"DEBIAN_FRONTEND": "noninteractive"},
                    apt_options=[
                        "--allow-downgrades",
                        '-o Dpkg::Options::="--force-confdef"',
                        '-o Dpkg::Options::="--force-confold"',
                    ],
                )
            except exceptions.UbuntuProError:
                progress.emit(
                    "info",
                    messages.FIPS_PACKAGE_NOT_AVAILABLE.format(
                        service=self.title, pkg=pkg
                    ),
                )

    def auto_upgrade_all_on_enable(self) -> bool:
        install_all_updates_override = util.is_config_value_true(
            config=self.cfg.cfg, path_to_value="features.fips_auto_upgrade_all"
        )
        # jammy onward automatically uses new auto-upgrade logic
        hardcoded_release = system.get_release_info().series in {
            "xenial",
            "bionic",
            "focal",
        }
        return install_all_updates_override or not hardcoded_release

    def install_all_available_fips_upgrades(
        self, progress: api.ProgressWrapper
    ):
        to_upgrade = [
            package.name
            for package in apt.get_installed_packages_with_uninstalled_candidate_in_origin(  # noqa: E501
                self.origin
            )
        ]

        # See GH #3390
        if system.get_release_info().series == "jammy":
            to_upgrade.append("openssl-fips-module-3")

        to_upgrade.sort()
        if len(to_upgrade) > 0:
            try:
                progress.emit(
                    "info",
                    messages.INSTALLING_PACKAGES.format(
                        packages=" ".join(to_upgrade)
                    ),
                )
                self.unhold_packages(to_upgrade)
                apt.run_apt_install_command(
                    packages=to_upgrade,
                    override_env_vars={"DEBIAN_FRONTEND": "noninteractive"},
                    apt_options=[
                        "--allow-downgrades",
                        '-o Dpkg::Options::="--force-confdef"',
                        '-o Dpkg::Options::="--force-confold"',
                    ],
                )
            except exceptions.UbuntuProError:
                progress.emit("info", messages.FIPS_PACKAGES_UPGRADE_FAILURE)

    def install_packages(
        self,
        progress: api.ProgressWrapper,
        package_list: Optional[List[str]] = None,
        cleanup_on_failure: bool = True,
    ) -> None:
        """Install contract recommended packages for the entitlement.

        :param package_list: Optional package list to use instead of
            self.packages.
        :param cleanup_on_failure: Cleanup apt files if apt install fails.
        """
        # We need to guarantee that the metapackage is installed.
        # While the other packages should still be installed, if they
        # fail, we should not block the enable operation.
        mandatory_packages = self.packages
        if mandatory_packages:
            super().install_packages(
                progress,
                package_list=mandatory_packages,
            )
        else:
            # then this won't get printed by install_packages, so do it here
            # instead
            progress.progress(
                messages.INSTALLING_SERVICE_PACKAGES.format(title=self.title)
            )

        if self.auto_upgrade_all_on_enable():
            self.install_all_available_fips_upgrades(progress)
        else:
            self.hardcoded_install_conditional_packages(progress)

        if self._check_for_reboot():
            notices.add(
                Notice.FIPS_SYSTEM_REBOOT_REQUIRED,
            )

    def _check_for_reboot(self) -> bool:
        """Check if system needs to be rebooted because of this service."""
        return system.should_reboot()

    def _check_for_reboot_msg(
        self, operation: str, silent: bool = False
    ) -> None:
        """Check if user should be alerted that a reboot must be performed.

        @param operation: The operation being executed.
        @param silent: Boolean set True to silence print/log of messages
        """
        reboot_required = self._check_for_reboot()
        event.needs_reboot(reboot_required)
        if reboot_required:
            if not silent:
                event.info(
                    messages.ENABLE_REBOOT_REQUIRED_TMPL.format(
                        operation=operation
                    )
                )
            if operation == "disable operation":
                notices.add(
                    Notice.FIPS_DISABLE_REBOOT_REQUIRED,
                )

    def _allow_fips_on_cloud_instance(
        self, series: str, cloud_id: str
    ) -> bool:
        """Return False when FIPS is allowed on this cloud and series.

        On Xenial GCP there will be no cloud-optimized kernel so
        block default ubuntu-fips enable. This can be overridden in
        config with features.allow_xenial_fips_on_cloud.

        GCP doesn't yet have a cloud-optimized kernel or metapackage so
        block enable of fips if the contract does not specify ubuntu-gcp-fips.
        This also can be overridden in config with
        features.allow_default_fips_metapackage_on_gcp.

        :return: False when this cloud, series or config override allows FIPS.
        """
        if cloud_id == "gce":
            if util.is_config_value_true(
                config=self.cfg.cfg,
                path_to_value="features.allow_default_fips_metapackage_on_gcp",
            ):
                return True

            # GCE only has FIPS support for bionic and focal machines
            if series in ("bionic", "focal"):
                return True

            return bool("ubuntu-gcp-fips" in super().packages)

        return True

    @property
    def static_affordances(self) -> Tuple[StaticAffordance, ...]:
        cloud_titles = {"aws": "an AWS", "azure": "an Azure", "gce": "a GCP"}
        cloud_id, _ = get_cloud_type()
        if cloud_id is None:
            cloud_id = ""

        series = system.get_release_info().series
        blocked_message = messages.FIPS_BLOCK_ON_CLOUD.format(
            series=series.title(), cloud=cloud_titles.get(cloud_id)
        )
        return (
            (
                blocked_message,
                lambda: self._allow_fips_on_cloud_instance(series, cloud_id),
                True,
            ),
        )

    @property
    def packages(self) -> List[str]:
        if system.is_container():
            return []
        return super().packages

    def application_status(
        self,
    ) -> Tuple[ApplicationStatus, Optional[messages.NamedMessage]]:
        super_status, super_msg = super().application_status()

        if system.is_container() and not system.should_reboot():
            notices.remove(
                Notice.FIPS_SYSTEM_REBOOT_REQUIRED,
            )
            return super_status, super_msg

        if os.path.exists(self.FIPS_PROC_FILE):
            # We are now only removing the notice if there is no reboot
            # required information regarding the fips metapackage we install.
            if not system.should_reboot(set(self.packages)):
                notices.remove(
                    Notice.FIPS_SYSTEM_REBOOT_REQUIRED,
                )

            if system.load_file(self.FIPS_PROC_FILE).strip() == "1":
                notices.remove(
                    Notice.FIPS_MANUAL_DISABLE_URL,
                )
                return super_status, super_msg
            else:
                notices.add(
                    Notice.FIPS_MANUAL_DISABLE_URL,
                )
                return (
                    ApplicationStatus.DISABLED,
                    messages.FIPS_PROC_FILE_ERROR.format(
                        file_name=self.FIPS_PROC_FILE
                    ),
                )

        if super_status != ApplicationStatus.ENABLED:
            return super_status, super_msg

        installed_packages = apt.get_installed_packages_names()
        missing = []
        # Should be a single package (ubuntu-fips and its cloud variants)
        # or no package at all for containers
        for package in self.packages:
            if package not in installed_packages:
                missing.append(package)

        if missing:
            return (
                ApplicationStatus.WARNING,
                messages.FIPS_PACKAGES_NOT_INSTALLED.format(
                    packages=" ".join(missing), service=self.name
                ),
            )

        return (
            ApplicationStatus.ENABLED,
            messages.FIPS_REBOOT_REQUIRED,
        )

    def remove_packages(self) -> None:
        """Remove fips meta package to disable the service.

        FIPS meta-package will unset grub config options which will deactivate
        FIPS on any related packages.
        """
        installed_packages = set(apt.get_installed_packages_names())
        fips_metapackage = set(self.packages).difference(
            set(self.conditional_packages)
        )
        remove_packages = fips_metapackage.intersection(installed_packages)
        if remove_packages:
            apt.remove_packages(
                list(fips_metapackage),
                messages.DISABLE_FAILED_TMPL.format(title=self.title),
            )

    def _perform_enable(self, progress: api.ProgressWrapper) -> bool:
        if super()._perform_enable(progress):
            notices.remove(
                Notice.WRONG_FIPS_METAPACKAGE_ON_CLOUD,
            )
            notices.remove(Notice.FIPS_REBOOT_REQUIRED)
            notices.remove(Notice.FIPS_DISABLE_REBOOT_REQUIRED)
            return True

        return False

    def _perform_disable(self, progress: api.ProgressWrapper) -> bool:
        if super()._perform_disable(progress):
            if self._check_for_reboot():
                notices.add(
                    Notice.FIPS_DISABLE_REBOOT_REQUIRED,
                )
            return True

        return False

    def unhold_packages(self, package_names):
        cmd = ["apt-mark", "showholds"]
        holds = apt.run_apt_command(
            cmd,
            messages.EXECUTING_COMMAND_FAILED.format(command=" ".join(cmd)),
        )
        unholds = []
        for hold in holds.splitlines():
            if hold in package_names:
                unholds.append(hold)
        if unholds:
            unhold_cmd = ["apt-mark", "unhold"] + unholds
            holds = apt.run_apt_command(
                unhold_cmd,
                messages.EXECUTING_COMMAND_FAILED.format(
                    command=" ".join(unhold_cmd)
                ),
            )

    def setup_apt_config(self, progress: api.ProgressWrapper) -> None:
        """Setup apt config based on the resourceToken and directives.

        FIPS-specifically handle apt-mark unhold

        :raise UbuntuProError: on failure to setup any aspect of this apt
           configuration
        """
        self.unhold_packages(self.fips_pro_package_holds)
        super().setup_apt_config(progress)


class FIPSEntitlement(FIPSCommonEntitlement):
    name = "fips"
    title = messages.FIPS_TITLE
    description = messages.FIPS_DESCRIPTION
    help_text = messages.FIPS_HELP_TEXT
    origin = "UbuntuFIPS"
    pre_enable_msg = messages.PROMPT_FIPS_PRE_ENABLE

    @property
    def incompatible_services(self) -> Tuple[EntitlementWithMessage, ...]:
        from uaclient.entitlements.livepatch import LivepatchEntitlement
        from uaclient.entitlements.realtime import RealtimeKernelEntitlement

        return (
            EntitlementWithMessage(
                LivepatchEntitlement, messages.LIVEPATCH_INVALIDATES_FIPS
            ),
            EntitlementWithMessage(
                FIPSUpdatesEntitlement, messages.FIPS_UPDATES_INVALIDATES_FIPS
            ),
            EntitlementWithMessage(
                RealtimeKernelEntitlement, messages.REALTIME_FIPS_INCOMPATIBLE
            ),
        )

    @property
    def static_affordances(self) -> Tuple[StaticAffordance, ...]:
        static_affordances = super().static_affordances

        fips_updates = FIPSUpdatesEntitlement(cfg=self.cfg)
        disabled_status = ApplicationStatus.DISABLED
        is_fips_updates_enabled = bool(
            fips_updates.application_status()[0] != disabled_status
        )

        services_once_enabled_obj = services_once_enabled_file.read()
        fips_updates_once_enabled = (
            services_once_enabled_obj.fips_updates
            if services_once_enabled_obj
            else False
        )

        return static_affordances + (
            (
                messages.FIPS_ERROR_WHEN_FIPS_UPDATES_ENABLED.format(
                    fips=self.title, fips_updates=fips_updates.title
                ),
                lambda: is_fips_updates_enabled,
                False,
            ),
            (
                messages.FIPS_ERROR_WHEN_FIPS_UPDATES_ONCE_ENABLED.format(
                    fips=self.title, fips_updates=fips_updates.title
                ),
                lambda: fips_updates_once_enabled,
                False,
            ),
        )

    def _perform_enable(self, progress: api.ProgressWrapper) -> bool:
        cloud_type, error = get_cloud_type()
        if cloud_type is None and error == NoCloudTypeReason.CLOUD_ID_ERROR:
            LOG.warning(
                "Could not determine cloud, "
                "defaulting to generic FIPS package."
            )
            event.info(messages.FIPS_COULD_NOT_DETERMINE_CLOUD_DEFAULT_PACKAGE)
        if super()._perform_enable(progress):
            notices.remove(
                Notice.FIPS_INSTALL_OUT_OF_DATE,
            )
            return True

        return False


class FIPSUpdatesEntitlement(FIPSCommonEntitlement):
    name = "fips-updates"
    title = messages.FIPS_UPDATES_TITLE
    origin = "UbuntuFIPSUpdates"
    description = messages.FIPS_UPDATES_DESCRIPTION
    help_text = messages.FIPS_UPDATES_HELP_TEXT
    pre_enable_msg = messages.PROMPT_FIPS_UPDATES_PRE_ENABLE

    @property
    def incompatible_services(self) -> Tuple[EntitlementWithMessage, ...]:
        from uaclient.entitlements.realtime import RealtimeKernelEntitlement

        return (
            EntitlementWithMessage(
                FIPSEntitlement, messages.FIPS_INVALIDATES_FIPS_UPDATES
            ),
            EntitlementWithMessage(
                RealtimeKernelEntitlement,
                messages.REALTIME_FIPS_UPDATES_INCOMPATIBLE,
            ),
        )

    def _perform_enable(self, progress: api.ProgressWrapper) -> bool:
        if super()._perform_enable(progress=progress):
            services_once_enabled_file.write(
                ServicesOnceEnabledData(fips_updates=True)
            )
            return True

        return False


class FIPSPreviewEntitlement(FIPSEntitlement):
    name = "fips-preview"
    title = messages.FIPS_PREVIEW_TITLE
    description = messages.FIPS_PREVIEW_DESCRIPTION
    help_text = messages.FIPS_PREVIEW_HELP_TEXT
    origin = "UbuntuFIPSPreview"
    pre_enable_msg = messages.PROMPT_FIPS_PREVIEW_PRE_ENABLE
    repo_key_file = "ubuntu-pro-fips-preview.gpg"

    @property
    def incompatible_services(self) -> Tuple[EntitlementWithMessage, ...]:
        return super().incompatible_services + (
            EntitlementWithMessage(
                FIPSEntitlement, messages.FIPS_INVALIDATES_FIPS_UPDATES
            ),
        )

    def _allow_fips_on_cloud_instance(
        self, series: str, cloud_id: str
    ) -> bool:
        # For fips-preview, we should not block the service
        # if there is no FIPS cloud-optimized kernel. That is
        # because this service is intended as an early access for
        # FIPS service, so users should be aware of the risks
        return True



import contextlib
import io
import json
from textwrap import dedent

import mock
import pytest

from uaclient import event_logger, exceptions, lock, messages
from uaclient.cli import main_error_handler
from uaclient.cli.detach import detach_command
from uaclient.testing.fakes import FakeContractClient

M_PATH = "uaclient.cli.detach."


def entitlement_cls_mock_factory(can_disable, name=None):
    m_instance = mock.MagicMock()
    m_instance.enabled_variant = None
    m_instance.can_disable.return_value = (can_disable, None)
    m_instance.disable.return_value = (can_disable, None)
    type(m_instance).dependent_services = mock.PropertyMock(return_value=())
    if name:
        type(m_instance).name = mock.PropertyMock(return_value=name)

    return m_instance


@mock.patch(M_PATH + "util.prompt_for_confirmation", return_value=True)
class TestActionDetach:
    @mock.patch(M_PATH + "util.we_are_currently_root", return_value=False)
    def test_non_root_users_are_rejected(
        self,
        m_we_are_currently_root,
        _m_prompt,
        FakeConfig,
        fake_machine_token_file,
        event,
        capsys,
    ):
        """Check that a UID != 0 will receive a message and exit non-zero"""
        m_we_are_currently_root.return_value = False
        args = mock.MagicMock()

        cfg = FakeConfig()
        fake_machine_token_file.attached = True
        with pytest.raises(exceptions.NonRootUserError):
            detach_command.action(args, cfg=cfg)

        with pytest.raises(SystemExit):
            with mock.patch.object(
                event, "_event_logger_mode", event_logger.EventLoggerMode.JSON
            ):
                main_error_handler(detach_command.action)(args, cfg)

        expected_message = messages.E_NONROOT_USER
        expected = {
            "_schema_version": event_logger.JSON_SCHEMA_VERSION,
            "result": "failure",
            "errors": [
                {
                    "message": expected_message.msg,
                    "message_code": expected_message.name,
                    "service": None,
                    "type": "system",
                }
            ],
            "failed_services": [],
            "needs_reboot": False,
            "processed_services": [],
            "warnings": [],
        }
        assert expected == json.loads(capsys.readouterr()[0])

    def test_unattached_error_message(
        self, _m_prompt, FakeConfig, capsys, event
    ):
        """Check that root user gets unattached message."""

        cfg = FakeConfig()
        args = mock.MagicMock()
        with pytest.raises(exceptions.UnattachedError) as err:
            detach_command.action(args, cfg=cfg)
        assert messages.E_UNATTACHED.msg == err.value.msg

        with pytest.raises(SystemExit):
            with mock.patch.object(
                event, "_event_logger_mode", event_logger.EventLoggerMode.JSON
            ):
                main_error_handler(detach_command.action)(args, cfg)

        expected_message = messages.E_UNATTACHED
        expected = {
            "_schema_version": event_logger.JSON_SCHEMA_VERSION,
            "result": "failure",
            "errors": [
                {
                    "message": expected_message.msg,
                    "message_code": expected_message.name,
                    "service": None,
                    "type": "system",
                }
            ],
            "failed_services": [],
            "needs_reboot": False,
            "processed_services": [],
            "warnings": [],
        }
        assert expected == json.loads(capsys.readouterr()[0])

    @mock.patch("uaclient.lock.check_lock_info")
    @mock.patch("time.sleep")
    @mock.patch("uaclient.system.subp")
    def test_lock_file_exists(
        self,
        m_subp,
        m_sleep,
        m_check_lock_info,
        m_prompt,
        FakeConfig,
        fake_machine_token_file,
        capsys,
        event,
    ):
        """Check when an operation holds a lock file, detach cannot run."""
        cfg = FakeConfig()
        fake_machine_token_file.attached = True
        args = mock.MagicMock()
        m_check_lock_info.return_value = (123, "pro enable")

        with pytest.raises(exceptions.LockHeldError) as err:
            detach_command.action(args, cfg=cfg)

        assert 12 == m_check_lock_info.call_count
        expected_error_msg = messages.E_LOCK_HELD_ERROR.format(
            lock_request="pro detach", lock_holder="pro enable", pid="123"
        )
        assert expected_error_msg.msg == err.value.msg

        with pytest.raises(SystemExit):
            with mock.patch.object(
                event, "_event_logger_mode", event_logger.EventLoggerMode.JSON
            ):
                main_error_handler(detach_command.action)(args, cfg)

        expected = {
            "_schema_version": event_logger.JSON_SCHEMA_VERSION,
            "result": "failure",
            "errors": [
                {
                    "additional_info": {
                        "lock_holder": "pro enable",
                        "lock_request": "pro detach",
                        "pid": 123,
                    },
                    "message": expected_error_msg.msg,
                    "message_code": expected_error_msg.name,
                    "service": None,
                    "type": "system",
                }
            ],
            "failed_services": [],
            "needs_reboot": False,
            "processed_services": [],
            "warnings": [],
        }
        assert expected == json.loads(capsys.readouterr()[0])

    @pytest.mark.parametrize(
        "prompt_response,assume_yes,expect_disable",
        [(True, False, True), (False, False, False), (True, True, True)],
    )
    @mock.patch("uaclient.files.state_files.delete_state_files")
    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch("uaclient.contract.UAContractClient")
    @mock.patch(M_PATH + "update_motd_messages")
    @mock.patch(M_PATH + "entitlements.entitlements_disable_order")
    @mock.patch(M_PATH + "entitlements.entitlement_factory")
    def test_entitlements_disabled_appropriately(
        self,
        m_ent_factory,
        m_disable_order,
        m_update_apt_and_motd_msgs,
        m_client,
        _m_check_lock_info,
        _m_delete_state_files,
        m_prompt,
        prompt_response,
        assume_yes,
        expect_disable,
        FakeConfig,
        fake_machine_token_file,
        event,
        capsys,
    ):
        # The three parameters:
        #   prompt_response: the user's response to the prompt
        #   assume_yes: the value of the --assume-yes flag in the args passed
        #               to the action
        #   expect_disable: whether or not the enabled entitlement is expected
        #                   to be disabled by the action

        cfg = FakeConfig()
        fake_machine_token_file.attached = True
        fake_client = FakeContractClient(cfg)
        m_client.return_value = fake_client

        m_prompt.return_value = prompt_response

        m_disable_order.return_value = ["test"]
        m_ent = entitlement_cls_mock_factory(True, name="test")
        m_ent_factory.return_value = m_ent

        args = mock.MagicMock(assume_yes=assume_yes)
        with mock.patch.object(lock, "lock_data_file"):
            return_code = detach_command.action(args, cfg=cfg)

        assert [
            mock.call(ignore_dependent_services=True)
        ] == m_ent.can_disable.call_args_list

        if expect_disable:
            assert [mock.call(mock.ANY)] == m_ent.disable.call_args_list
            assert 0 == return_code
        else:
            assert 0 == m_ent.disable.call_count
            assert 1 == return_code
        assert [mock.call(assume_yes=assume_yes)] == m_prompt.call_args_list
        if expect_disable:
            assert [
                mock.call(cfg)
            ] == m_update_apt_and_motd_msgs.call_args_list

        fake_stdout = io.StringIO()
        # On json response, we will never prompt the user
        m_prompt.return_value = True
        with contextlib.redirect_stdout(fake_stdout):
            with mock.patch.object(
                event, "_event_logger_mode", event_logger.EventLoggerMode.JSON
            ):
                with mock.patch.object(lock, "lock_data_file"):
                    main_error_handler(detach_command.action)(args, cfg)

        expected = {
            "_schema_version": event_logger.JSON_SCHEMA_VERSION,
            "result": "success",
            "errors": [],
            "failed_services": [],
            "needs_reboot": False,
            "processed_services": ["test"],
            "warnings": [],
        }
        assert expected == json.loads(fake_stdout.getvalue())

    @mock.patch("uaclient.files.state_files.delete_state_files")
    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch(M_PATH + "cli_util._is_attached")
    @mock.patch(M_PATH + "entitlements.entitlements_disable_order")
    @mock.patch(M_PATH + "update_motd_messages")
    def test_correct_message_emitted(
        self,
        m_update_apt_and_motd_msgs,
        m_disable_order,
        m_is_attached,
        _m_check_lock_info,
        m_delete_state_files,
        _m_prompt,
        capsys,
        tmpdir,
    ):
        m_disable_order.return_value = []
        m_is_attached.return_value = mock.MagicMock(
            is_attached=True,
            contract_status="active",
            contract_remaining_days=100,
        )

        m_cfg = mock.MagicMock()
        with mock.patch.object(lock, "lock_data_file"):
            detach_command.action(mock.MagicMock(), m_cfg)

        out, _err = capsys.readouterr()
        assert messages.DETACH_SUCCESS + "\n" == out
        assert [mock.call(m_cfg)] == m_update_apt_and_motd_msgs.call_args_list
        assert [mock.call()] == m_delete_state_files.call_args_list

    @mock.patch("uaclient.files.state_files.delete_state_files")
    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch(M_PATH + "cli_util._is_attached")
    @mock.patch(M_PATH + "entitlements.entitlements_disable_order")
    @mock.patch(M_PATH + "update_motd_messages")
    def test_returns_zero(
        self,
        m_update_apt_and_motd_msgs,
        m_disable_order,
        m_is_attached,
        _m_check_lock_info,
        m_delete_state_files,
        _m_prompt,
        tmpdir,
    ):
        m_disable_order.return_value = []
        m_is_attached.return_value = mock.MagicMock(
            is_attached=True,
            contract_status="active",
            contract_remaining_days=100,
        )

        m_cfg = mock.MagicMock()
        with mock.patch.object(lock, "lock_data_file"):
            ret = detach_command.action(mock.MagicMock(), m_cfg)

        assert 0 == ret
        assert [mock.call()] == m_delete_state_files.call_args_list
        assert [mock.call(m_cfg)] == m_update_apt_and_motd_msgs.call_args_list

    @pytest.mark.parametrize(
        "classes,disable_order,expected_message,disabled_services",
        [
            (
                [
                    entitlement_cls_mock_factory(True, name="ent1"),
                    entitlement_cls_mock_factory(False, name="ent2"),
                    entitlement_cls_mock_factory(True, name="ent3"),
                ],
                ["ent1", "ent2", "ent3"],
                dedent(
                    """\
                    Detach will disable the following services:
                        ent1
                        ent3"""
                ),
                ["ent1", "ent3"],
            ),
            (
                [
                    entitlement_cls_mock_factory(True, name="ent1"),
                    entitlement_cls_mock_factory(False, name="ent2"),
                ],
                ["ent1", "ent2"],
                dedent(
                    """\
                    Detach will disable the following service:
                        ent1"""
                ),
                ["ent1"],
            ),
        ],
    )
    @mock.patch("uaclient.files.state_files.delete_state_files")
    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch(M_PATH + "cli_util._is_attached")
    @mock.patch(M_PATH + "update_motd_messages")
    @mock.patch("uaclient.entitlements.entitlement_factory")
    @mock.patch(M_PATH + "entitlements.entitlements_disable_order")
    def test_informational_message_emitted(
        self,
        m_disable_order,
        m_ent_factory,
        m_update_apt_and_motd_msgs,
        m_is_attached,
        _m_check_lock_info,
        _m_delete_state_files,
        _m_prompt,
        capsys,
        classes,
        disable_order,
        expected_message,
        disabled_services,
        tmpdir,
        FakeConfig,
        fake_machine_token_file,
        event,
    ):
        m_ent_factory.side_effect = classes
        m_disable_order.return_value = disable_order
        m_is_attached.return_value = mock.MagicMock(
            is_attached=True,
            contract_status="active",
            contract_remaining_days=100,
        )

        m_cfg = mock.MagicMock()
        args = mock.MagicMock()

        with mock.patch.object(lock, "lock_data_file"):
            detach_command.action(args, m_cfg)

        out, _err = capsys.readouterr()

        assert expected_message in out
        assert [mock.call(m_cfg)] == m_update_apt_and_motd_msgs.call_args_list

        fake_stdout = io.StringIO()
        m_ent_factory.side_effect = classes
        fake_machine_token_file.attached = True
        with contextlib.redirect_stdout(fake_stdout):
            with mock.patch.object(
                event, "_event_logger_mode", event_logger.EventLoggerMode.JSON
            ):
                with mock.patch.object(lock, "lock_data_file"):
                    main_error_handler(detach_command.action)(
                        args, FakeConfig()
                    )

        expected = {
            "_schema_version": event_logger.JSON_SCHEMA_VERSION,
            "result": "success",
            "errors": [],
            "failed_services": [],
            "needs_reboot": False,
            "processed_services": disabled_services,
            "warnings": [],
        }
        assert expected == json.loads(fake_stdout.getvalue())



import mock
import pytest

from lib.daemon import (
    WAIT_FOR_CLOUD_CONFIG_POLL_TIMES,
    WAIT_FOR_CLOUD_CONFIG_SLEEP_TIME,
    _wait_for_cloud_config,
)


class TestWaitForCloudConfig:
    @pytest.mark.parametrize(
        [
            "active_state_side_effect",
            "expected_sleep_calls",
        ],
        (
            # not activating
            (
                ["active"] * 2,
                [],
            ),
            # inactive (all cloud-init)
            (
                ["inactive"] * 2,
                [],
            ),
            (
                [None] * 2,
                [],
            ),
            # cloud-config activating, then finishes
            # cloud-init is active
            (
                (["activating", "active"] * 11) + ["active"] * 2,
                [mock.call(WAIT_FOR_CLOUD_CONFIG_SLEEP_TIME)] * 11,
            ),
            (
                (["activating", "active"] * 11) + ["failed"] + ["active"],
                [mock.call(WAIT_FOR_CLOUD_CONFIG_SLEEP_TIME)] * 11,
            ),
            # inactive cloud-config, active cloud-init
            (
                (["inactive", "active"] * 6)
                + (["activating", "active"] * 5)
                + ["active"] * 2,
                [mock.call(WAIT_FOR_CLOUD_CONFIG_SLEEP_TIME)] * 11,
            ),
            # inactive cloud-config, activating cloud-init
            (
                (["inactive", "activating"] * 2)
                + (["inactive", "active"] * 4)
                + (["activating", "active"] * 5)
                + ["active"] * 2,
                [mock.call(WAIT_FOR_CLOUD_CONFIG_SLEEP_TIME)] * 11,
            ),
            # still activating after polling maximum times
            (
                ["activating"] * (WAIT_FOR_CLOUD_CONFIG_POLL_TIMES + 1000),
                [mock.call(WAIT_FOR_CLOUD_CONFIG_SLEEP_TIME)]
                * WAIT_FOR_CLOUD_CONFIG_POLL_TIMES,
            ),
        ),
    )
    @mock.patch("lib.daemon.time.sleep")
    @mock.patch("lib.daemon.system.get_systemd_unit_active_state")
    def test_wait_for_cloud_config(
        self,
        m_get_systemd_unit_active_state,
        m_sleep,
        active_state_side_effect,
        expected_sleep_calls,
    ):
        m_get_systemd_unit_active_state.side_effect = active_state_side_effect
        _wait_for_cloud_config()
        assert m_sleep.call_args_list == expected_sleep_calls



"""Tests related to uaclient.apt module."""

import glob
import os
import stat
import subprocess
from textwrap import dedent
from typing import List, Optional

import mock
import pytest

from uaclient import exceptions, messages, system
from uaclient.apt import (
    APT_AUTH_COMMENT,
    APT_CONFIG_GLOBAL_PROXY_HTTP,
    APT_CONFIG_GLOBAL_PROXY_HTTPS,
    APT_HELPER_TIMEOUT,
    APT_KEYS_DIR,
    APT_PROXY_CONF_FILE,
    APT_PROXY_CONFIG_HEADER,
    APT_RETRIES,
    KEYRINGS_DIR,
    SERIES_NOT_USING_DEB822,
    PreserveAptCfg,
    _ensure_esm_cache_structure,
    add_apt_auth_conf_entry,
    add_auth_apt_repo,
    add_ppa_pinning,
    assert_valid_apt_credentials,
    find_apt_list_files,
    get_apt_cache_policy,
    get_apt_cache_time,
    get_apt_config_values,
    get_installed_packages_by_origin,
    get_installed_packages_names,
    get_installed_packages_with_uninstalled_candidate_in_origin,
    get_pkg_candidate_version,
    get_remote_versions_for_package,
    get_system_sources_file,
    is_installed,
    remove_apt_list_files,
    remove_auth_apt_repo,
    remove_repo_from_apt_auth_file,
    run_apt_update_command,
    setup_apt_proxy,
    update_esm_caches,
    update_sources_list,
)
from uaclient.entitlements.base import UAEntitlement
from uaclient.entitlements.entitlement_status import ApplicationStatus
from uaclient.entitlements.repo import RepoEntitlement
from uaclient.testing import fakes

POST_INSTALL_APT_CACHE_NO_UPDATES = """
-32768 https://esm.ubuntu.com/ubuntu/ {0}-updates/main amd64 Packages
     release v=14.04,o={1},a={0}-updates,n={0},l=UbuntuESM,c=main
     origin esm.ubuntu.com
"""

APT_LIST_RETURN_STRING = """\
"Listing... Done
a/release, now 1.2+3 arch123 [i,a]
b/release-updates, now 1.2+3 arch123 [i,a]
"""


def mock_origin(
    component: str, archive: str, origin: str, site: str
) -> mock.MagicMock:
    mock_origin = mock.MagicMock()
    mock_origin.component = component
    mock_origin.archive = archive
    mock_origin.origin = origin
    mock_origin.site = site
    return [mock_origin, 0]


def mock_version(
    version: str,
    origin_list: List[mock.MagicMock] = [],
    size: int = 1,
) -> mock.MagicMock:
    mock_version = mock.MagicMock()
    mock_version.__gt__ = lambda self, other: self.ver_str > other.ver_str
    mock_version.ver_str = version
    mock_version.file_list = origin_list
    mock_version.size = size
    return mock_version


def mock_package(
    name: str,
    installed_version: Optional[mock.MagicMock] = None,
    other_versions: List[mock.MagicMock] = [],
):
    mock_package = mock.MagicMock()
    mock_package.name = name
    mock_package.version_list = []

    mock_package.current_ver = None
    if installed_version:
        mock_package.current_ver = installed_version
        installed_version.parent_pkg = mock_package
        mock_package.version_list.append(installed_version)

    for version in other_versions:
        version.parent_pkg = mock_package
        mock_package.version_list.append(version)

    return mock_package


class TestAddPPAPinning:
    @mock.patch("uaclient.system.get_release_info")
    def test_write_apt_pin_file_to_apt_preferences(
        self, m_get_release_info, tmpdir
    ):
        """Write proper apt pin file to specified apt_preference_file."""
        m_get_release_info.return_value = mock.MagicMock(series="xenial")
        pref_file = tmpdir.join("preffile").strpath
        assert None is add_ppa_pinning(
            pref_file,
            repo_url="http://fakerepo",
            origin="MYORIG",
            priority=1003,
        )
        expected_pref = dedent(
            """\
            Package: *
            Pin: release o=MYORIG
            Pin-Priority: 1003\n"""
        )
        assert expected_pref == system.load_file(pref_file)


class TestFindAptListFilesFromRepoSeries:
    @mock.patch("uaclient.system.subp")
    def test_find_all_apt_list_files_from_apt_config_key(self, m_subp, tmpdir):
        """Find all matching apt list files from apt-config dir."""
        m_subp.return_value = ("key='{}'".format(tmpdir.strpath), "")
        repo_url = "http://c.com/fips-updates/"
        _protocol, repo_path = repo_url.split("://")
        prefix = repo_path.rstrip("/").replace("/", "_")
        paths = sorted(
            [
                tmpdir.join(prefix + "_dists_nomatch").strpath,
                tmpdir.join(prefix + "_dists_xenial_InRelease").strpath,
                tmpdir.join(
                    prefix + "_dists_xenial_main_binary-amd64_Packages"
                ).strpath,
            ]
        )
        for path in paths:
            system.write_file(path, "")

        assert paths[1:] == find_apt_list_files(repo_url, "xenial")


class TestRemoveAptListFiles:
    @mock.patch("uaclient.system.subp")
    def test_remove_all_apt_list_files_from_apt_config_key(
        self, m_subp, tmpdir
    ):
        """Remove all matching apt list files from apt-config dir."""
        m_subp.return_value = ("key='{}'".format(tmpdir.strpath), "")
        repo_url = "http://c.com/fips-updates/"
        _protocol, repo_path = repo_url.split("://")
        prefix = repo_path.rstrip("/").replace("/", "_")
        nomatch_file = tmpdir.join(prefix + "_dists_nomatch").strpath
        paths = [
            nomatch_file,
            tmpdir.join(prefix + "_dists_xenial_InRelease").strpath,
            tmpdir.join(
                prefix + "_dists_xenial_main_binary-amd64_Packages"
            ).strpath,
        ]
        for path in paths:
            system.write_file(path, "")

        assert None is remove_apt_list_files(repo_url, "xenial")
        assert [nomatch_file] == glob.glob("{}/*".format(tmpdir.strpath))


class TestValidAptCredentials:
    @mock.patch("uaclient.system.subp")
    @mock.patch("os.path.exists", return_value=False)
    def test_passes_when_missing_apt_helper(self, m_exists, m_subp):
        """When apt-helper tool is absent perform no validation."""
        assert None is assert_valid_apt_credentials(
            repo_url="http://fakerepo", username="username", password="pass"
        )
        expected_calls = [mock.call("/usr/lib/apt/apt-helper")]
        assert expected_calls == m_exists.call_args_list
        assert 0 == m_subp.call_count

    @mock.patch("uaclient.apt.tempfile.TemporaryDirectory")
    @mock.patch("uaclient.system.subp")
    @mock.patch("uaclient.apt.os.path.exists", return_value=True)
    def test_passes_on_valid_creds(
        self, m_exists, m_subp, m_temporary_directory
    ):
        """Succeed when apt-helper succeeds in authenticating to repo."""
        m_temporary_directory.return_value.__enter__.return_value = (
            "/does/not/exist"
        )
        # Success apt-helper response
        m_subp.return_value = "Get:1 https://fakerepo\nFetched 285 B in 1s", ""

        assert None is assert_valid_apt_credentials(
            repo_url="http://fakerepo", username="user", password="pwd"
        )
        exists_calls = [mock.call("/usr/lib/apt/apt-helper")]
        assert exists_calls == m_exists.call_args_list
        expected_path = os.path.join(
            m_temporary_directory.return_value.__enter__.return_value,
            "apt-helper-output",
        )
        apt_helper_call = mock.call(
            [
                "/usr/lib/apt/apt-helper",
                "download-file",
                "http://user:pwd@fakerepo/pool/",
                expected_path,
            ],
            timeout=60,
            retry_sleeps=APT_RETRIES,
        )
        assert [apt_helper_call] == m_subp.call_args_list

    @pytest.mark.parametrize(
        "exit_code,stderr,error_msg",
        (
            (
                1,
                "something broke",
                (
                    "Unexpected APT error.\n"
                    "Failed running command 'apt-helper ' [exit(1)]. Message: "
                    "something broke\n"
                    "See /var/log/ubuntu-advantage.log"
                ),
            ),
            (
                100,
                "E: Failed to fetch ... HttpError401 on xenial",
                "Invalid APT credentials provided for http://fakerepo",
            ),
            (
                100,
                "E: Failed to fetch ... 401 Unauthorized on xenial",
                "Invalid APT credentials provided for http://fakerepo",
            ),
            (
                100,
                "E: Failed to fetch ... Connection timed out",
                "Timeout trying to access APT repository at http://fakerepo",
            ),
        ),
    )
    @mock.patch("uaclient.apt.tempfile.TemporaryDirectory")
    @mock.patch("uaclient.system.subp")
    @mock.patch("uaclient.apt.os.path.exists", return_value=True)
    def test_errors_on_process_execution_errors(
        self,
        m_exists,
        m_subp,
        m_temporary_directory,
        exit_code,
        stderr,
        error_msg,
    ):
        """Raise the appropriate user facing error from apt-helper failure."""
        m_temporary_directory.return_value.__enter__.return_value = (
            "/does/not/exist"
        )
        # Failure apt-helper response
        m_subp.side_effect = exceptions.ProcessExecutionError(
            cmd="apt-helper ",
            exit_code=exit_code,
            stdout="Err:1...",
            stderr=stderr,
        )

        with pytest.raises(exceptions.UbuntuProError) as excinfo:
            assert_valid_apt_credentials(
                repo_url="http://fakerepo", username="user", password="pwd"
            )
        assert error_msg == str(excinfo.value)
        exists_calls = [mock.call("/usr/lib/apt/apt-helper")]
        assert exists_calls == m_exists.call_args_list
        expected_path = os.path.join(
            m_temporary_directory.return_value.__enter__.return_value,
            "apt-helper-output",
        )
        apt_helper_call = mock.call(
            [
                "/usr/lib/apt/apt-helper",
                "download-file",
                "http://user:pwd@fakerepo/pool/",
                expected_path,
            ],
            timeout=60,
            retry_sleeps=APT_RETRIES,
        )
        assert [apt_helper_call] == m_subp.call_args_list

    @mock.patch("uaclient.apt.tempfile.TemporaryDirectory")
    @mock.patch("uaclient.system.subp")
    @mock.patch("uaclient.apt.os.path.exists", return_value=True)
    def test_errors_on_apt_helper_process_timeout(
        self, m_exists, m_subp, m_temporary_directory
    ):
        """Raise the appropriate user facing error from apt-helper timeout."""
        m_temporary_directory.return_value.__enter__.return_value = (
            "/does/not/exist"
        )
        # Failure apt-helper response
        m_subp.side_effect = subprocess.TimeoutExpired(
            "something timed out", timeout=1000000
        )
        with pytest.raises(exceptions.UbuntuProError) as excinfo:
            assert_valid_apt_credentials(
                repo_url="http://fakerepo", username="user", password="pwd"
            )
        error_msg = (
            "Cannot validate credentials for APT repo. Timeout"
            " after {} seconds trying to reach fakerepo.".format(
                APT_HELPER_TIMEOUT
            )
        )
        assert error_msg == excinfo.value.msg
        exists_calls = [mock.call("/usr/lib/apt/apt-helper")]
        assert exists_calls == m_exists.call_args_list
        expected_path = os.path.join(
            m_temporary_directory.return_value.__enter__.return_value,
            "apt-helper-output",
        )
        apt_helper_call = mock.call(
            [
                "/usr/lib/apt/apt-helper",
                "download-file",
                "http://user:pwd@fakerepo/pool/",
                expected_path,
            ],
            timeout=APT_HELPER_TIMEOUT,
            retry_sleeps=APT_RETRIES,
        )
        assert [apt_helper_call] == m_subp.call_args_list


class TestAddAuthAptRepo:
    @pytest.mark.parametrize("series", ("xenial", "noble"))
    @mock.patch("uaclient.apt.gpg.export_gpg_key")
    @mock.patch("uaclient.system.subp")
    @mock.patch("uaclient.apt.get_apt_auth_file_from_apt_config")
    @mock.patch("uaclient.apt.assert_valid_apt_credentials")
    @mock.patch("uaclient.system.get_release_info")
    def test_add_auth_apt_repo_writes_sources_file(
        self,
        m_get_release_info,
        m_valid_creds,
        m_get_apt_auth_file,
        m_subp,
        m_gpg_export,
        series,
        tmpdir,
    ):
        """Write a properly configured sources file to repo_filename."""
        repo_file = tmpdir.join("repo.conf").strpath
        auth_file = tmpdir.join("auth.conf").strpath
        m_get_apt_auth_file.return_value = auth_file
        m_subp.return_value = "500 esm.canonical.com...", ""  # apt policy
        m_get_release_info.return_value = mock.MagicMock(series=series)

        add_auth_apt_repo(
            repo_filename=repo_file,
            repo_url="http://fakerepo",
            credentials="mycreds",
            suites=(series,),
            keyring_file="keyring",
            snapshot_urls=[],
        )

        if series in SERIES_NOT_USING_DEB822:
            expected_content = (
                "deb http://fakerepo {series} main\n"
                "# deb-src http://fakerepo {series} main\n"
            ).format(series=series)
            src_keyfile = os.path.join(KEYRINGS_DIR, "keyring")
            dest_keyfile = os.path.join(APT_KEYS_DIR, "keyring")
            gpg_export_calls = [mock.call(src_keyfile, dest_keyfile)]
        else:
            expected_content = (
                "# Written by ubuntu-pro-client\n"
                "Types: deb\n"
                "URIs: http://fakerepo\n"
                "Suites: {series}\n"
                "Components: main\n"
                "Signed-By: /usr/share/keyrings/keyring\n"
            ).format(series=series)
            gpg_export_calls = []

        assert expected_content == system.load_file(repo_file)
        assert gpg_export_calls == m_gpg_export.call_args_list

    @pytest.mark.parametrize("series", ("xenial", "noble"))
    @mock.patch("uaclient.apt.gpg.export_gpg_key")
    @mock.patch("uaclient.system.subp")
    @mock.patch("uaclient.apt.get_apt_auth_file_from_apt_config")
    @mock.patch("uaclient.apt.assert_valid_apt_credentials")
    @mock.patch("uaclient.system.get_release_info")
    def test_add_auth_apt_repo_ignores_suites_not_matching_series(
        self,
        m_get_release_info,
        m_valid_creds,
        m_get_apt_auth_file,
        m_subp,
        m_gpg_export,
        series,
        tmpdir,
    ):
        """Skip any apt suites that don't match the current series."""
        repo_file = tmpdir.join("repo.conf").strpath
        auth_file = tmpdir.join("auth.conf").strpath
        m_get_apt_auth_file.return_value = auth_file
        # apt policy with series-updates enabled
        stdout = dedent(
            """\
            500 http://archive.ubuntu.com/ {series}-updates/main amd64 \
                        Packages
                release v=XX.04,o=Ubuntu,a={series}-updates,n={series},\
                        l=Ubuntu,c=main""".format(
                series=series
            )
        )
        m_subp.return_value = stdout, ""
        m_get_release_info.return_value = mock.MagicMock(series=series)

        add_auth_apt_repo(
            repo_filename=repo_file,
            repo_url="http://fakerepo",
            credentials="mycreds",
            suites=(
                "{}-one".format(series),
                "{}-updates".format(series),
                "trusty-gone",
            ),
            keyring_file="keyring",
            snapshot_urls=[],
        )

        if series in SERIES_NOT_USING_DEB822:
            expected_content = dedent(
                """\
                deb http://fakerepo {series}-one main
                # deb-src http://fakerepo {series}-one main
                deb http://fakerepo {series}-updates main
                # deb-src http://fakerepo {series}-updates main
            """
            ).format(series=series)
        else:
            expected_content = dedent(
                """\
                # Written by ubuntu-pro-client
                Types: deb
                URIs: http://fakerepo
                Suites: {series}-one {series}-updates
                Components: main
                Signed-By: /usr/share/keyrings/keyring
            """
            ).format(series=series)

        assert expected_content == system.load_file(repo_file)

    @pytest.mark.parametrize("series", ("xenial", "noble"))
    @mock.patch("uaclient.apt.gpg.export_gpg_key")
    @mock.patch("uaclient.system.subp")
    @mock.patch("uaclient.apt.get_apt_auth_file_from_apt_config")
    @mock.patch("uaclient.apt.assert_valid_apt_credentials")
    @mock.patch("uaclient.system.get_release_info")
    def test_add_auth_apt_repo_comments_updates_suites_on_non_update_machine(
        self,
        m_get_release_info,
        m_valid_creds,
        m_get_apt_auth_file,
        m_subp,
        m_gpg_export,
        series,
        tmpdir,
    ):
        """Skip any apt suites that don't match the current series."""
        repo_file = tmpdir.join("repo.conf").strpath
        auth_file = tmpdir.join("auth.conf").strpath
        m_get_apt_auth_file.return_value = auth_file
        # apt policy without series-updates enabled
        m_subp.return_value = (
            POST_INSTALL_APT_CACHE_NO_UPDATES.format(series, "test-origin"),
            "",
        )
        m_get_release_info.return_value = mock.MagicMock(series=series)

        add_auth_apt_repo(
            repo_filename=repo_file,
            repo_url="http://fakerepo",
            credentials="mycreds",
            suites=(
                "{}-one".format(series),
                "{}-updates".format(series),
                "trusty-gone",
            ),
            keyring_file="keyring",
            snapshot_urls=[],
            check_updates_pocket=True,
        )

        if series in SERIES_NOT_USING_DEB822:
            expected_content = dedent(
                """\
                deb http://fakerepo {series}-one main
                # deb-src http://fakerepo {series}-one main
                # deb http://fakerepo {series}-updates main
                # deb-src http://fakerepo {series}-updates main
            """
            ).format(series=series)
        else:
            expected_content = dedent(
                """\
                # Written by ubuntu-pro-client
                Types: deb
                URIs: http://fakerepo
                Suites: {series}-one
                Components: main
                Signed-By: /usr/share/keyrings/keyring
            """
            ).format(series=series)

        assert expected_content == system.load_file(repo_file)

    @mock.patch("uaclient.apt.gpg.export_gpg_key")
    @mock.patch("uaclient.system.subp")
    @mock.patch("uaclient.apt.get_apt_auth_file_from_apt_config")
    @mock.patch("uaclient.apt.assert_valid_apt_credentials")
    @mock.patch(
        "uaclient.system.get_release_info",
        return_value=mock.MagicMock(series="xenial"),
    )
    def test_add_auth_apt_repo_writes_username_password_to_auth_file(
        self,
        m_get_release_info,
        m_valid_creds,
        m_get_apt_auth_file,
        m_subp,
        m_gpg_export,
        tmpdir,
    ):
        """Write apt authentication file when credentials are user:pwd."""
        repo_file = tmpdir.join("repo.conf").strpath
        auth_file = tmpdir.join("auth.conf").strpath
        m_get_apt_auth_file.return_value = auth_file
        m_subp.return_value = "500 esm.canonical.com...", ""  # apt policy

        add_auth_apt_repo(
            repo_filename=repo_file,
            repo_url="http://fakerepo",
            credentials="user:password",
            suites=("xenial",),
            keyring_file="keyring",
            snapshot_urls=[],
        )

        expected_content = (
            "machine fakerepo/ login user password password"
            "{}\n".format(APT_AUTH_COMMENT)
        )
        assert expected_content == system.load_file(auth_file)

    @mock.patch("uaclient.apt.gpg.export_gpg_key")
    @mock.patch("uaclient.system.subp")
    @mock.patch("uaclient.apt.get_apt_auth_file_from_apt_config")
    @mock.patch("uaclient.apt.assert_valid_apt_credentials")
    @mock.patch(
        "uaclient.system.get_release_info",
        return_value=mock.MagicMock(series="xenial"),
    )
    def test_add_auth_apt_repo_writes_bearer_resource_token_to_auth_file(
        self,
        m_get_release_info,
        m_valid_creds,
        m_get_apt_auth_file,
        m_subp,
        m_gpg_export,
        tmpdir,
    ):
        """Write apt authentication file when credentials are bearer token."""
        repo_file = tmpdir.join("repo.conf").strpath
        auth_file = tmpdir.join("auth.conf").strpath
        m_get_apt_auth_file.return_value = auth_file
        m_subp.return_value = "500 esm.canonical.com...", ""  # apt policy

        add_auth_apt_repo(
            repo_filename=repo_file,
            repo_url="http://fakerepo/",
            credentials="SOMELONGTOKEN",
            suites=("xenia",),
            keyring_file="keyring",
            snapshot_urls=[],
        )

        expected_content = (
            "machine fakerepo/ login bearer password"
            " SOMELONGTOKEN{}\n".format(APT_AUTH_COMMENT)
        )
        assert expected_content == system.load_file(auth_file)


class TestAddAptAuthConfEntry:
    @mock.patch("uaclient.apt.get_apt_auth_file_from_apt_config")
    def test_replaces_old_credentials_with_new(
        self, m_get_apt_auth_file, tmpdir
    ):
        """Replace old credentials for this repo_url on the same line."""
        auth_file = tmpdir.join("auth.conf").strpath
        system.write_file(
            auth_file,
            dedent(
                """\
            machine fakerepo1/ login me password password1
            machine fakerepo/ login old password oldpassword
            machine fakerepo2/ login other password otherpass
        """
            ),
        )

        m_get_apt_auth_file.return_value = auth_file

        add_apt_auth_conf_entry(
            login="newlogin", password="newpass", repo_url="http://fakerepo/"
        )

        content_template = dedent(
            """\
            machine fakerepo1/ login me password password1
            machine fakerepo/ login newlogin password newpass{}
            machine fakerepo2/ login other password otherpass
        """
        )
        expected_content = content_template.format(APT_AUTH_COMMENT)
        assert expected_content == system.load_file(auth_file)

    @mock.patch("uaclient.apt.get_apt_auth_file_from_apt_config")
    def test_insert_repo_subroutes_before_existing_repo_basepath(
        self, m_get_apt_auth_file, tmpdir
    ):
        """Insert new repo_url before first matching url base path."""
        auth_file = tmpdir.join("auth.conf").strpath
        system.write_file(
            auth_file,
            dedent(
                """\
            machine fakerepo1/ login me password password1
            machine fakerepo/ login old password oldpassword
            machine fakerepo2/ login other password otherpass
        """
            ),
        )

        m_get_apt_auth_file.return_value = auth_file

        add_apt_auth_conf_entry(
            login="new",
            password="newpass",
            repo_url="http://fakerepo/subroute",
        )

        content_template = dedent(
            """\
            machine fakerepo1/ login me password password1
            machine fakerepo/subroute/ login new password newpass{}
            machine fakerepo/ login old password oldpassword
            machine fakerepo2/ login other password otherpass
        """
        )
        expected_content = content_template.format(APT_AUTH_COMMENT)
        assert expected_content == system.load_file(auth_file)


class TestCleanAptFiles:
    @pytest.fixture(params=[RepoEntitlement, UAEntitlement])
    def mock_apt_entitlement(self, request, tmpdir):
        # Set up our tmpdir with some fake list files
        entitlement_name = "test_ent"
        repo_tmpl = tmpdir.join("source-{name}").strpath
        pref_tmpl = tmpdir.join("pref-{name}").strpath

        class TestRepo(request.param):
            name = entitlement_name
            repo_file_tmpl = repo_tmpl
            repo_pref_file_tmpl = pref_tmpl
            is_repo = request.param == RepoEntitlement

        for series in ["acidic", "base"]:
            source_name = repo_tmpl.format(name=entitlement_name)
            pref_name = pref_tmpl.format(name=entitlement_name)

            with open(source_name, "w") as f:
                f.write("")

            with open(pref_name, "w") as f:
                f.write("")

        return TestRepo


@pytest.fixture(params=(mock.sentinel.default, None, "some_string"))
def remove_auth_apt_repo_kwargs(request):
    """
    Parameterized fixture to generate all permutations of kwargs we need

    Note that this tests three states for keyring_file: using the default,
    explicitly passing None and explicitly passing a string.
    """
    keyring_file = request.param
    kwargs = {}
    if keyring_file != mock.sentinel.default:
        kwargs["keyring_file"] = keyring_file
    return kwargs


@mock.patch("uaclient.apt.system.subp")
@mock.patch("uaclient.apt.remove_repo_from_apt_auth_file")
@mock.patch("uaclient.apt.system.ensure_file_absent")
class TestRemoveAuthAptRepo:
    def test_repo_file_deleted(
        self,
        m_ensure_file_absent,
        _m_remove_repo,
        _m_subp,
        remove_auth_apt_repo_kwargs,
    ):
        """Ensure that repo_filename is deleted, regardless of other params."""
        repo_filename = "/etc/apt/sources.list.d/pro-repofile.list"
        repo_url = mock.sentinel.url

        remove_auth_apt_repo(
            repo_filename, repo_url, [], **remove_auth_apt_repo_kwargs
        )

        assert mock.call(repo_filename) in m_ensure_file_absent.call_args_list

    def test_old_repo_file_deleted_when_deb822(
        self,
        m_ensure_file_absent,
        _m_remove_repo,
        _m_subp,
        remove_auth_apt_repo_kwargs,
    ):
        repo_filename = "/etc/apt/sources.list.d/pro-repofile.sources"
        repo_url = mock.sentinel.url

        remove_auth_apt_repo(
            repo_filename, repo_url, [], **remove_auth_apt_repo_kwargs
        )

        assert mock.call(repo_filename) in m_ensure_file_absent.call_args_list
        assert (
            mock.call("/etc/apt/sources.list.d/pro-repofile.list")
            in m_ensure_file_absent.call_args_list
        )

    def test_remove_from_auth_file_called(
        self,
        _m_ensure_file_absent,
        m_remove_repo,
        _m_subp,
        remove_auth_apt_repo_kwargs,
    ):
        """Ensure that remove_repo_from_apt_auth_file is called."""
        repo_filename = "/etc/apt/sources.list.d/pro-repofile.list"
        repo_url = mock.sentinel.url

        remove_auth_apt_repo(
            repo_filename, repo_url, [], **remove_auth_apt_repo_kwargs
        )

        assert mock.call(repo_url) in m_remove_repo.call_args_list

    def test_keyring_file_deleted_if_given(
        self,
        m_ensure_file_absent,
        _m_remove_repo,
        _m_subp,
        remove_auth_apt_repo_kwargs,
    ):
        """We should always delete the keyring file if it is given"""
        repo_filename = "/etc/apt/sources.list.d/pro-repofile.list"
        repo_url = mock.sentinel.url

        remove_auth_apt_repo(
            repo_filename, repo_url, [], **remove_auth_apt_repo_kwargs
        )

        keyring_file = remove_auth_apt_repo_kwargs.get("keyring_file")
        if keyring_file:
            assert (
                mock.call(os.path.join(APT_KEYS_DIR, keyring_file))
                in m_ensure_file_absent.call_args_list
            )
        else:
            assert (
                mock.call(keyring_file)
                not in m_ensure_file_absent.call_args_list
            )


class TestRemoveRepoFromAptAuthFile:
    @mock.patch("uaclient.system.ensure_file_absent")
    @mock.patch("uaclient.apt.system.write_file")
    @mock.patch("uaclient.apt.get_apt_auth_file_from_apt_config")
    def test_auth_file_doesnt_exist_means_we_dont_remove_or_write_it(
        self, m_get_apt_auth_file, m_write_file, m_ensure_file_absent, tmpdir
    ):
        """If the auth file doesn't exist, we shouldn't do anything to it"""
        m_get_apt_auth_file.return_value = tmpdir.join("nonexistent").strpath

        remove_repo_from_apt_auth_file("http://url")

        assert 0 == m_write_file.call_count
        assert 0 == m_ensure_file_absent.call_count

    @pytest.mark.parametrize("trailing_slash", (True, False))
    @pytest.mark.parametrize(
        "repo_url,auth_file_content",
        (
            ("http://url1", b""),
            ("http://url2", b"machine url2/ login trailing content"),
            ("http://url3", b"machine url3/ login"),
            ("http://url4", b"leading content machine url4/ login"),
            (
                "http://url4",
                b"leading content machine url4/ login trailing content",
            ),
        ),
    )
    @mock.patch("uaclient.system.ensure_file_absent")
    @mock.patch("uaclient.apt.system.write_file")
    @mock.patch("uaclient.apt.get_apt_auth_file_from_apt_config")
    def test_file_removal(
        self,
        m_get_apt_auth_file,
        m_write_file,
        m_ensure_file_absent,
        tmpdir,
        trailing_slash,
        repo_url,
        auth_file_content,
    ):
        """Check that auth file is rm'd if empty or contains just our line"""
        auth_file = tmpdir.join("auth_file")
        auth_file.write(auth_file_content, "wb")
        m_get_apt_auth_file.return_value = auth_file.strpath

        remove_repo_from_apt_auth_file(
            repo_url + ("" if not trailing_slash else "/")
        )

        assert 0 == m_write_file.call_count
        assert [
            mock.call(auth_file.strpath)
        ] == m_ensure_file_absent.call_args_list

    @pytest.mark.parametrize("trailing_slash", (True, False))
    @pytest.mark.parametrize(
        "repo_url,before_content,after_content",
        (
            (
                "http://url1",
                b"should not be changed",
                b"should not be changed",
            ),
            (
                "http://url1",
                b"line before\nmachine url1/ login",
                b"line before",
            ),
            ("http://url1", b"machine url1/ login\nline after", b"line after"),
            (
                "http://url1",
                b"line before\nmachine url1/ login\nline after",
                b"line before\nline after",
            ),
            (
                "http://url1",
                b"unicode \xe2\x98\x83\nmachine url1/ login",
                b"unicode \xe2\x98\x83",
            ),
        ),
    )
    @mock.patch("uaclient.system.ensure_file_absent")
    @mock.patch("uaclient.apt.get_apt_auth_file_from_apt_config")
    def test_file_rewrite(
        self,
        m_get_apt_auth_file,
        m_ensure_file_absent,
        tmpdir,
        repo_url,
        before_content,
        after_content,
        trailing_slash,
    ):
        """Check that auth file is rewritten to only exclude our line"""
        auth_file = tmpdir.join("auth_file")
        auth_file.write(before_content, "wb")
        m_get_apt_auth_file.return_value = auth_file.strpath

        remove_repo_from_apt_auth_file(
            repo_url + ("" if not trailing_slash else "/")
        )

        assert 0 == m_ensure_file_absent.call_count
        assert 0o600 == stat.S_IMODE(os.lstat(auth_file.strpath).st_mode)
        assert after_content == auth_file.read("rb")


class TestGetInstalledPackagesNames:
    @pytest.mark.parametrize(
        [
            "cache_packages",
            "expected_result",
        ],
        [
            (
                [
                    mock_package("one", mock_version("1", [])),
                    mock_package("two"),  # not installed
                    mock_package("three", mock_version("1", [])),
                    mock_package("four", mock_version("1", [])),
                ],
                ["one", "three", "four"],
            )
        ],
    )
    @mock.patch("uaclient.apt.get_apt_pkg_cache")
    def test_get_installed_packages_names(
        self,
        m_apt_cache,
        cache_packages,
        expected_result,
    ):
        m_apt_cache.return_value.packages = cache_packages
        assert expected_result == get_installed_packages_names()


class TestRunAptCommand:
    @pytest.mark.parametrize(
        "error_list, output_list",
        (
            (
                [
                    "E: The repository 't1 404' does not have a Release.",
                    "W: Failed to fetch t1/dists/ 404 Not found.",
                    "E: The repository 't2 404' does not have a Release.\n",
                ],
                (
                    "APT update failed.",
                    (
                        "APT update failed to read APT config "
                        "for the following:"
                    ),
                    "- t1",
                    "- t2",
                ),
            ),
            (
                ["E: The repository 't1 404' does not have a Release file."],
                (
                    "APT update failed.",
                    (
                        "APT update failed to read APT config "
                        "for the following:"
                    ),
                    "- t1",
                ),
            ),
            (
                [
                    "W: Failed to fetch t1/dists Not Found [IP: 127.0.0.1]\n",
                    "E: Some index files failed to download.\n",
                ],
                (
                    "APT update failed.",
                    (
                        "APT update failed to read APT config "
                        "for the following:"
                    ),
                    "- t1",
                ),
            ),
        ),
    )
    @mock.patch("uaclient.apt.system.subp")
    def test_run_apt_command_with_invalid_repositories(
        self, m_subp, error_list, output_list
    ):
        error_msg = "\n".join(error_list)

        m_subp.side_effect = exceptions.ProcessExecutionError(
            cmd="apt update", stderr=error_msg
        )

        with pytest.raises(exceptions.UbuntuProError) as excinfo:
            run_apt_update_command()

        expected_message = "\n".join(output_list)
        assert expected_message == excinfo.value.msg

    @mock.patch("uaclient.apt.system.subp")
    def test_run_update_command_clean_apt_cache_policy_cache(self, m_subp):
        m_subp.side_effect = [
            ("policy1", ""),
            ("update", ""),
            ("policy2", ""),
        ]

        assert "policy1" == get_apt_cache_policy()
        # Confirming that caching is happening
        assert "policy1" == get_apt_cache_policy()

        run_apt_update_command()

        # Confirm cache was cleared
        assert "policy2" == get_apt_cache_policy()
        get_apt_cache_policy.cache_clear()

    @mock.patch("uaclient.apt.system.subp")
    def test_failed_run_update_command_clean_apt_cache_policy_cache(
        self, m_subp
    ):
        m_subp.side_effect = [
            ("policy1", ""),
            fakes.FakeUbuntuProError(),
            ("policy2", ""),
        ]

        assert "policy1" == get_apt_cache_policy()
        # Confirming that caching is happening
        assert "policy1" == get_apt_cache_policy()

        with pytest.raises(exceptions.UbuntuProError):
            run_apt_update_command()

        # Confirm cache was cleared
        assert "policy2" == get_apt_cache_policy()
        get_apt_cache_policy.cache_clear()


class TestAptProxyConfig:
    @pytest.mark.parametrize(
        "kwargs, expected_remove_calls, expected_write_calls, expected_out",
        [
            ({}, [mock.call(APT_PROXY_CONF_FILE)], [], ""),
            (
                {"http_proxy": "mock_http_proxy"},
                [],
                [
                    mock.call(
                        APT_PROXY_CONF_FILE,
                        APT_PROXY_CONFIG_HEADER
                        + APT_CONFIG_GLOBAL_PROXY_HTTP.format(
                            proxy_url="mock_http_proxy"
                        ),
                    )
                ],
                messages.SETTING_SERVICE_PROXY_SCOPE.format(scope="global"),
            ),
            (
                {"https_proxy": "mock_https_proxy"},
                [],
                [
                    mock.call(
                        APT_PROXY_CONF_FILE,
                        APT_PROXY_CONFIG_HEADER
                        + APT_CONFIG_GLOBAL_PROXY_HTTPS.format(
                            proxy_url="mock_https_proxy"
                        ),
                    )
                ],
                messages.SETTING_SERVICE_PROXY_SCOPE.format(scope="global"),
            ),
            (
                {
                    "http_proxy": "mock_http_proxy",
                    "https_proxy": "mock_https_proxy",
                },
                [],
                [
                    mock.call(
                        APT_PROXY_CONF_FILE,
                        APT_PROXY_CONFIG_HEADER
                        + APT_CONFIG_GLOBAL_PROXY_HTTP.format(
                            proxy_url="mock_http_proxy"
                        )
                        + APT_CONFIG_GLOBAL_PROXY_HTTPS.format(
                            proxy_url="mock_https_proxy"
                        ),
                    )
                ],
                messages.SETTING_SERVICE_PROXY_SCOPE.format(scope="global"),
            ),
        ],
    )
    @mock.patch("uaclient.system.write_file")
    @mock.patch("uaclient.system.ensure_file_absent")
    def test_setup_apt_proxy_config(
        self,
        m_ensure_file_absent,
        m_util_write_file,
        kwargs,
        expected_remove_calls,
        expected_write_calls,
        expected_out,
        capsys,
        event,
    ):
        setup_apt_proxy(**kwargs)
        assert expected_remove_calls == m_ensure_file_absent.call_args_list
        assert expected_write_calls == m_util_write_file.call_args_list
        out, err = capsys.readouterr()
        assert expected_out == out.strip()
        assert "" == err


class TestAptIsInstalled:
    @pytest.mark.parametrize(
        "expected,installed_pkgs",
        (
            (True, ("foo", "test", "bar")),
            (False, ("foo", "bar")),
        ),
    )
    @mock.patch("uaclient.apt.get_installed_packages_names")
    def test_is_installed_pkgs(
        self, m_get_installed_pkgs, expected, installed_pkgs
    ):
        m_get_installed_pkgs.return_value = installed_pkgs
        assert expected == is_installed("test")


class TestAptCache:
    @pytest.mark.parametrize(
        "file_exists,expected", ((True, 1.23), (False, None))
    )
    @mock.patch("os.stat")
    @mock.patch("os.path.exists")
    def test_get_apt_cache_time(self, m_exists, m_stat, file_exists, expected):
        m_stat.return_value.st_mtime = 1.23
        m_exists.return_value = file_exists
        assert expected == get_apt_cache_time()

    @pytest.mark.parametrize(
        "is_lts,cache_call_list",
        ((True, [mock.call()]), (False, [])),
    )
    @pytest.mark.parametrize(
        "apps_status", (ApplicationStatus.ENABLED, ApplicationStatus.DISABLED)
    )
    @pytest.mark.parametrize(
        "infra_status", (ApplicationStatus.ENABLED, ApplicationStatus.DISABLED)
    )
    @pytest.mark.parametrize("is_esm", (True, False))
    @pytest.mark.parametrize("can_enable_infra", ("yes", "no"))
    @pytest.mark.parametrize("can_enable_apps", ("yes", "no"))
    @mock.patch(
        "uaclient.files.state_files.status_cache_file.read", return_value=None
    )
    @mock.patch("uaclient.entitlements.esm.ESMAppsEntitlement")
    @mock.patch("uaclient.entitlements.esm.ESMInfraEntitlement")
    @mock.patch("uaclient.apt.system.is_current_series_lts")
    @mock.patch("uaclient.apt.system.is_current_series_active_esm")
    @mock.patch("uaclient.apt.get_esm_apt_pkg_cache")
    @mock.patch("uaclient.apt._ensure_esm_cache_structure")
    @mock.patch("uaclient.actions.status")
    @mock.patch("apt_pkg.config")
    @mock.patch("apt_pkg.init_config")
    def test_update_esm_caches_based_on_lts(
        self,
        _m_apt_pkg_init_config,
        _m_apt_pkg_config,
        m_status,
        _m_ensure_structure,
        m_esm_cache,
        m_is_esm,
        m_is_lts,
        m_infra_entitlement,
        m_apps_entitlement,
        m_status_cache_file_read,
        is_lts,
        cache_call_list,
        apps_status,
        infra_status,
        is_esm,
        can_enable_infra,
        can_enable_apps,
        FakeConfig,
    ):
        m_status.return_value = (
            {
                "services": [
                    {"name": "esm-apps", "available": can_enable_apps},
                    {"name": "esm-infra", "available": can_enable_infra},
                ]
            },
            0,
        )

        m_is_esm.return_value = is_esm
        m_is_lts.return_value = is_lts

        m_infra = mock.MagicMock()
        m_apps = mock.MagicMock()

        m_infra.application_status.return_value = (
            infra_status,
            "",
        )

        m_apps.application_status.return_value = (
            apps_status,
            "",
        )

        m_infra_entitlement.return_value = m_infra
        m_apps_entitlement.return_value = m_apps

        infra_setup_repo_count = 0
        apps_setup_repo_count = 0
        infra_disable_repo_count = 0
        apps_disable_repo_count = 0
        status_count = 0

        if is_lts:
            status_count = 1
            if (
                apps_status == ApplicationStatus.DISABLED
                and can_enable_apps == "yes"
            ):
                apps_setup_repo_count = 1
            else:
                apps_disable_repo_count = 1

            if (
                infra_status == ApplicationStatus.DISABLED
                and is_esm
                and can_enable_infra == "yes"
            ):
                infra_setup_repo_count = 1
            elif is_esm:
                infra_disable_repo_count = 1

        cfg = FakeConfig()
        update_esm_caches(cfg)

        assert status_count == m_status_cache_file_read.call_count
        assert m_esm_cache.call_args_list == cache_call_list
        assert (
            m_infra.setup_local_esm_repo.call_count == infra_setup_repo_count
        )
        assert m_apps.setup_local_esm_repo.call_count == apps_setup_repo_count
        assert (
            m_infra.disable_local_esm_repo.call_count
            == infra_disable_repo_count
        )
        assert (
            m_apps.disable_local_esm_repo.call_count == apps_disable_repo_count
        )
        assert m_status.call_count == status_count


class TestGetAptConfigValues:
    @mock.patch("uaclient.apt._get_apt_config")
    def test_apt_config_values(
        self,
        m_get_apt_config,
    ):
        m_dict = mock.MagicMock()
        m_get_apt_config.return_value = m_dict

        m_dict.get.side_effect = ["", "foo", "bar", ""]
        m_dict.value_list.side_effect = [
            "",
            ["test1", "test2"],
        ]

        expected_return = {
            "val1": None,
            "val2": "foo",
            "val3": "bar",
            "val4": ["test1", "test2"],
        }

        assert expected_return == get_apt_config_values(
            ["val1", "val2", "val3", "val4"]
        )


class TestPreserveAptCfg:
    def test_apt_config_is_preserved(self):
        class AptDict(dict):
            def set(self, key, value):
                super().__setitem__(key, value)

        apt_cfg = AptDict()
        apt_cfg["test"] = 1
        apt_cfg["test1"] = [1, 2, 3]
        apt_cfg["test2"] = {"foo": "bar"}

        def apt_func():
            apt_cfg["test"] = 3
            apt_cfg["test1"] = [3, 2, 1]
            apt_cfg["test2"] = {"foo": "test"}
            return apt_cfg

        with mock.patch("apt_pkg.config", apt_cfg):
            with PreserveAptCfg(apt_func):
                pass

        assert 1 == apt_cfg["test"]
        assert [1, 2, 3] == apt_cfg["test1"]
        assert {"foo": "bar"} == apt_cfg["test2"]


class TestGetPkgCandidateversion:
    @pytest.mark.parametrize("check_esm_cache", (True, False))
    @mock.patch("uaclient.apt.apt_pkg.DepCache")
    @mock.patch("uaclient.apt.get_apt_pkg_cache")
    @mock.patch("uaclient.apt.get_esm_apt_pkg_cache")
    def test_get_pkg_candidate_version(
        self,
        m_esm_cache,
        m_apt_cache,
        m_dep_cache,
        check_esm_cache,
    ):
        m_pkg = mock.MagicMock()
        m_apt_cache.return_value = {"pkg1": m_pkg}

        m_esm_pkg = mock.MagicMock()
        m_esm_cache.return_value = {"pkg1": m_esm_pkg}

        def depcache_candidate(package):
            if package == m_pkg:
                return mock.MagicMock(ver_str="1.2")
            if package == m_esm_pkg:
                return mock.MagicMock(ver_str="1.3~esm1")

        m_dep_cache.return_value.get_candidate_ver = depcache_candidate

        actual_value = get_pkg_candidate_version("pkg1", check_esm_cache)
        if not check_esm_cache:
            assert "1.2" == actual_value
        else:
            assert "1.3~esm1" == actual_value

    @mock.patch("uaclient.apt.apt_pkg.DepCache")
    @mock.patch("uaclient.apt.get_apt_pkg_cache")
    @mock.patch("uaclient.apt.get_esm_apt_pkg_cache")
    def test_get_pkg_candidate_version_when_esm_cache_fails(
        self,
        m_esm_cache,
        m_apt_cache,
        m_dep_cache,
    ):
        def depcache_candidate(_package):
            return mock.MagicMock(ver_str="1.2")

        m_dep_cache.return_value.get_candidate_ver = depcache_candidate

        m_apt_cache.return_value = {"pkg1": mock.MagicMock()}
        m_esm_cache.return_value = {}

        actual_value = get_pkg_candidate_version("pkg1", check_esm_cache=True)
        assert "1.2" == actual_value

    @mock.patch("uaclient.apt.get_apt_pkg_cache")
    def test_get_pkg_candidate_version_when_package_doesnt_exist(
        self,
        m_apt_cache,
    ):
        m_apt_cache.return_value = {}

        actual_value = get_pkg_candidate_version("pkg1")
        assert actual_value is None

    @mock.patch("uaclient.apt.apt_pkg.DepCache")
    @mock.patch("uaclient.apt.get_apt_pkg_cache")
    def test_get_pkg_candidate_version_when_candidate_doesnt_exist(
        self,
        m_apt_cache,
        m_dep_cache,
    ):
        def depcache_candidate(_package):
            return None

        m_dep_cache.return_value.get_candidate_ver = depcache_candidate
        m_apt_cache.return_value = {"pkg1": mock.MagicMock()}

        actual_value = get_pkg_candidate_version("pkg1")
        assert actual_value is None


class TestGetInstalledPackagesByOrigin:
    @mock.patch("uaclient.apt.get_apt_pkg_cache")
    def test_packages_by_origin(self, m_apt_cache):
        origin_a_security = mock_origin(
            "main", "series-security", "OriginA", ""
        )
        origin_a_updates = mock_origin("main", "series-updates", "OriginA", "")
        origin_b = mock_origin("main", "series", "OriginB", "")
        m_apt_cache.return_value.packages = [
            mock_package("not_installed"),
            mock_package("origin_a", mock_version("1.0", [origin_a_security])),
            mock_package("origin_b", mock_version("1.0", [origin_b])),
            mock_package(
                "both_origins",
                mock_version("1.0", [origin_a_updates, origin_b]),
            ),
            mock_package(
                "origin_a_multiple_entries",
                mock_version("1.0", [origin_a_security, origin_a_updates]),
            ),
            mock_package(
                "origin_b_with_origin_a_not_installed",
                mock_version("1.0", [origin_b]),
                [mock_version("1.1", [origin_a_security, origin_a_updates])],
            ),
        ]

        assert [
            "both_origins",
            "origin_a",
            "origin_a_multiple_entries",
        ] == sorted(
            [p.name for p in get_installed_packages_by_origin("OriginA")]
        )
        assert [
            "both_origins",
            "origin_b",
            "origin_b_with_origin_a_not_installed",
        ] == sorted(
            [p.name for p in get_installed_packages_by_origin("OriginB")]
        )


class TestGetInstalledPackagesWithUninstalledCandidateByOrigin:
    @pytest.mark.parametrize(
        [
            "origin",
            "cache_packages",
            "get_candidate_ver_side_effect",
            "expected_result",
        ],
        [
            (
                "OriginA",
                [
                    mock_package("one", mock_version("1", [])),
                    mock_package("two"),  # not installed
                    mock_package("three", mock_version("1", [])),
                    mock_package("four", mock_version("1", [])),
                ],
                [
                    mock_version(
                        "2", [mock_origin("main", "series", "OriginA", "")]
                    ),
                    mock_version(
                        "2", [mock_origin("main", "series", "OriginB", "")]
                    ),
                    mock_version(
                        "2", [mock_origin("main", "series", "OriginA", "")]
                    ),
                ],
                ["four", "one"],
            )
        ],
    )
    @mock.patch("uaclient.apt.apt_pkg.DepCache")
    @mock.patch("uaclient.apt.get_apt_pkg_cache")
    def test_get_installed_packages_with_uninstalled_candidate_by_origin(
        self,
        m_apt_cache,
        m_dep_cache,
        origin,
        cache_packages,
        get_candidate_ver_side_effect,
        expected_result,
    ):
        m_apt_cache.return_value.packages = cache_packages
        m_dep_cache.return_value.get_candidate_ver.side_effect = (
            get_candidate_ver_side_effect
        )
        assert expected_result == sorted(
            [
                p.name
                for p in get_installed_packages_with_uninstalled_candidate_in_origin(  # noqa: E501
                    origin
                )
            ]
        )


class TestGetRemoteVersionsForPackage:
    def test_get_remote_versions_for_package(self):
        assert ["0.9", "1.1"] == [
            v.ver_str
            for v in sorted(
                get_remote_versions_for_package(
                    mock_package(
                        "name",
                        installed_version=mock_version("1.0"),
                        other_versions=[
                            mock_version(
                                "0.9", [mock_origin("", "", "Origin1", "")]
                            ),
                            mock_version(
                                "1.1", [mock_origin("", "", "Origin2", "")]
                            ),
                        ],
                    )
                )
            )
        ]

    def test_no_remote_versions(self):
        assert [] == get_remote_versions_for_package(
            mock_package("name", installed_version=mock_version("1.0"))
        )

    def test_get_remote_versions_excluding_origin(self):
        assert ["0.9"] == [
            v.ver_str
            for v in sorted(
                get_remote_versions_for_package(
                    mock_package(
                        "name",
                        installed_version=mock_version(
                            "1.0", [mock_origin("now", "", "Origin1", "")]
                        ),
                        other_versions=[
                            mock_version(
                                "0.9", [mock_origin("", "", "Origin1", "")]
                            ),
                            mock_version(
                                "1.1", [mock_origin("", "", "Origin2", "")]
                            ),
                        ],
                    ),
                    exclude_origin="Origin2",
                )
            )
        ]


class TestUpdateSourcesList:
    @mock.patch("uaclient.apt.os.path.join")
    @mock.patch("apt_pkg.FileLock")
    @mock.patch("uaclient.apt.AcquireProgress")
    @mock.patch("apt_pkg.SourceList")
    @mock.patch("apt_pkg.config")
    @mock.patch("uaclient.apt.PreserveAptCfg")
    def test_update_sources_list(
        self,
        m_preserve_apt_cfg,
        m_apt_pkg_config,
        m_apt_pkg_source_list,
        m_acquire_progress,
        _m_file_lock,
        _m_join,
    ):
        update_sources_list("/sources.list")
        assert (
            mock.call("Dir::Etc::sourcelist", "/sources.list")
            in m_apt_pkg_config.set.call_args_list
        )
        assert (
            [
                mock.call(
                    m_acquire_progress.return_value,
                    m_apt_pkg_source_list.return_value,
                    0,
                )
            ]
            == m_preserve_apt_cfg.return_value.__enter__.return_value.update.call_args_list  # noqa: E501
        )


@mock.patch("uaclient.apt.system.ensure_folder_absent")
@mock.patch("uaclient.apt.system.create_file")
@mock.patch("uaclient.apt.os.makedirs")
class TestEsmCacheStructure:
    def test_noop_when_present(
        self, m_makedirs, m_create_file, m_folder_absent, tmpdir
    ):
        dir_path = tmpdir.strpath
        # Both mocks because ESM_APT_ROOTDIR can't deal with it
        with mock.patch("uaclient.apt.ESM_APT_ROOTDIR", dir_path), mock.patch(
            "uaclient.apt.ESM_BASIC_FILE_STRUCTURE",
            {
                "files": [
                    os.path.join(dir_path, "etc/apt/sources.list"),
                    os.path.join(dir_path, "var/lib/dpkg/status"),
                ],
                "folders": [
                    os.path.join(dir_path, "var/cache/apt/archives/partial"),
                    os.path.join(dir_path, "var/lib/apt/lists/partial"),
                ],
            },
        ):
            # Guess what: cannot use tmp_path because the fixture is not on
            # Xenial, and tmpdir does *not* support creating directories
            # including parents (makedirs style)
            tmpdir.mkdir("var")
            tmpdir.mkdir("var/cache")
            tmpdir.mkdir("var/cache/apt")
            tmpdir.mkdir("var/cache/apt/archives")
            tmpdir.mkdir("var/cache/apt/archives/partial")
            tmpdir.mkdir("var/lib")
            tmpdir.mkdir("var/lib/dpkg")
            # yes
            tmpdir.join("var/lib/dpkg/status").write("")
            tmpdir.mkdir("var/lib/apt")
            tmpdir.mkdir("var/lib/apt/lists")
            tmpdir.mkdir("var/lib/apt/lists/partial")
            tmpdir.mkdir("etc")
            tmpdir.mkdir("etc/apt")
            tmpdir.join("etc/apt/sources.list").write("")

            _ensure_esm_cache_structure()

        assert m_folder_absent.call_args_list == []
        assert m_create_file.call_args_list == []
        assert m_makedirs.call_args_list == []

    @pytest.mark.parametrize("create_partial", (False, True))
    def test_create_when_necessary(
        self,
        m_makedirs,
        m_create_file,
        m_folder_absent,
        create_partial,
        tmpdir,
    ):
        dir_path = tmpdir.strpath
        with mock.patch("uaclient.apt.ESM_APT_ROOTDIR", dir_path), mock.patch(
            "uaclient.apt.ESM_BASIC_FILE_STRUCTURE",
            {
                "files": [
                    os.path.join(dir_path, "etc/apt/sources.list"),
                    os.path.join(dir_path, "var/lib/dpkg/status"),
                ],
                "folders": [
                    os.path.join(dir_path, "var/cache/apt/archives/partial"),
                    os.path.join(dir_path, "var/lib/apt/lists/partial"),
                ],
            },
        ):
            if create_partial:
                tmpdir.mkdir("etc")
                tmpdir.mkdir("etc/apt")
                tmpdir.join("etc/apt/sources.list").write("")

            _ensure_esm_cache_structure()

        assert m_folder_absent.call_args_list == [mock.call(tmpdir)]
        assert m_create_file.call_args_list == [
            mock.call(tmpdir + "/etc/apt/sources.list"),
            mock.call(tmpdir + "/var/lib/dpkg/status"),
        ]
        assert m_makedirs.call_args_list == [
            mock.call(
                tmpdir + "/var/cache/apt/archives/partial",
                exist_ok=True,
                mode=0o755,
            ),
            mock.call(
                tmpdir + "/var/lib/apt/lists/partial",
                exist_ok=True,
                mode=0o755,
            ),
        ]


class TestGetSystemSourcesFile:
    @pytest.mark.parametrize(
        "new_source_exists,expected_return",
        (
            (True, "/etc/apt/sources.list.d/ubuntu.sources"),
            (False, "/etc/apt/sources.list"),
        ),
    )
    @mock.patch("uaclient.apt.os.path.exists")
    def test_get_system_sources_file(
        self, m_exists, new_source_exists, expected_return
    ):
        m_exists.return_value = new_source_exists
        assert get_system_sources_file() == expected_return



import json
import sys
from collections import OrderedDict
from typing import Any, Optional  # noqa: F401

from uaclient import exceptions, messages
from uaclient.api import AbstractProgress
from uaclient.api.api import call_api
from uaclient.cli.commands import ProArgument, ProArgumentGroup, ProCommand
from uaclient.cli.parser import HelpCategory


class CLIAPIProgress(AbstractProgress):
    def progress(
        self,
        *,
        total_steps: int,
        done_steps: int,
        previous_step_message: Optional[str],
        current_step_message: Optional[str]
    ):
        d = OrderedDict()  # type: OrderedDict[str, Any]
        d["total_steps"] = total_steps
        d["done_steps"] = done_steps
        d["previous_step_message"] = previous_step_message
        d["current_step_message"] = current_step_message
        print(json.dumps(d))


def action_api(args, *, cfg, **kwargs):
    if args.options and args.data:
        raise exceptions.CLIAPIOptionsXORData()

    if args.data and args.data == "-":
        if not sys.stdin.isatty():
            args.data = sys.stdin.read()

    if args.show_progress:
        progress = CLIAPIProgress()
    else:
        progress = None

    result = call_api(
        args.endpoint_path, args.options, args.data, cfg, progress
    )
    print(result.to_json())
    return 0 if result.result == "success" else 1


api_command = ProCommand(
    "api",
    help=messages.CLI_ROOT_API,
    description=messages.CLI_API_DESC,
    action=action_api,
    help_category=HelpCategory.OTHER,
    preserve_description=True,
    argument_groups=[
        ProArgumentGroup(
            arguments=[
                ProArgument(
                    "endpoint_path",
                    help=messages.CLI_API_ENDPOINT,
                    metavar="endpoint",
                ),
                ProArgument(
                    "--show-progress",
                    help=messages.CLI_API_SHOW_PROGRESS,
                    action="store_true",
                ),
                ProArgument(
                    "--args",
                    help=messages.CLI_API_ARGS,
                    dest="options",
                    default=[],
                    nargs="*",
                ),
                ProArgument(
                    "--data",
                    help=messages.CLI_API_DATA,
                    dest="data",
                    default="",
                ),
            ]
        )
    ],
)



from typing import List, Optional

from uaclient.data_types import (
    BoolDataValue,
    DataObject,
    DatetimeDataValue,
    Field,
    IntDataValue,
    StringDataValue,
    data_list,
)


class ActivityInfo(DataObject):
    fields = [
        Field("activityID", StringDataValue, False),
        Field("activityToken", StringDataValue, False),
        Field("activityPingInterval", IntDataValue, False),
        Field("resources", data_list(StringDataValue), False),
    ]

    def __init__(
        self,
        activityID: Optional[str],
        activityToken: Optional[str],
        activityPingInterval: Optional[int],
        resources: Optional[List[str]],
    ):
        self.activityID = activityID
        self.activityPingInterval = activityPingInterval
        self.activityToken = activityToken
        self.resources = resources


class AvailableResource(DataObject):
    fields = [
        Field("available", BoolDataValue, False),
        Field("name", StringDataValue, False),
        Field("description", StringDataValue, False),
        Field("presentedAs", StringDataValue, False),
    ]

    def __init__(
        self,
        available: Optional[bool],
        name: Optional[str],
        description: Optional[str],
        presentedAs: Optional[str],
    ):
        self.available = available
        self.name = name
        self.description = description
        self.presentedAs = presentedAs


class ExternalID(DataObject):
    fields = [
        Field("origin", StringDataValue, False),
        Field("IDs", data_list(StringDataValue), False),
    ]

    def __init__(self, origin: Optional[str], IDs: Optional[List[str]]):
        self.origin = origin
        self.IDs = IDs


class AccountInfo(DataObject):
    fields = [
        Field("name", StringDataValue, False),
        Field("id", StringDataValue, False),
        Field("createdAt", DatetimeDataValue, False),
        Field("type", StringDataValue, False),
        Field("userRoleOnAccount", StringDataValue, False),
        Field("externalAccountIDs", data_list(ExternalID), False),
    ]

    def __init__(
        self,
        name: Optional[str],
        id: Optional[str],
        createdAt: Optional[str],
        type: Optional[str],
        userRoleOnAccount: Optional[str],
        externalAccountIDs: Optional[List[str]],
    ):
        self.name = name
        self.id = id
        self.createdAt = createdAt
        self.type = type
        self.userRoleOnAccount = userRoleOnAccount
        self.externalAccountIDs = externalAccountIDs


class PlatformChecks(DataObject):
    fields = [
        Field("cpu_vendor_ids", data_list(StringDataValue), False),
    ]

    def __init__(
        self,
        cpu_vendor_ids: Optional[List[str]],
    ):
        self.cpu_vendor_ids = cpu_vendor_ids


class Affordances(DataObject):
    fields = [
        Field("architectures", data_list(StringDataValue), False),
        Field("presentedAs", StringDataValue, False),
        Field("series", data_list(StringDataValue), False),
        Field("kernelFlavors", data_list(StringDataValue), False),
        Field("minKernelVersion", StringDataValue, False),
        Field("tier", StringDataValue, False),
        Field("supportLevel", StringDataValue, False),
        Field("platformChecks", PlatformChecks, False),
    ]

    def __init__(
        self,
        architectures: Optional[List[str]],
        presentedAs: Optional[str],
        series: List[Optional[str]],
        kernelFlavors: List[Optional[str]],
        minKernelVersion: Optional[str],
        tier: Optional[str],
        supportLevel: Optional[str],
        platformChecks: Optional[PlatformChecks],
    ):
        self.architectures = architectures
        self.presentedAs = presentedAs
        self.series = series
        self.kernelFlavors = kernelFlavors
        self.minKernelVersion = minKernelVersion
        self.tier = tier
        self.supportLevel = supportLevel
        self.platformChecks = platformChecks


class Obligations(DataObject):
    fields = [
        Field("enableByDefault", BoolDataValue, False),
        Field("additionalProperties", StringDataValue, False),
    ]

    def __init__(
        self,
        enableByDefault: Optional[bool],
        additionalProperties: Optional[str],
    ):
        self.enableByDefault = enableByDefault
        self.additionalProperties = additionalProperties


class Directives(DataObject):
    fields = [
        Field("additionalPackages", data_list(StringDataValue), False),
        Field("aptURL", StringDataValue, False),
        Field("suites", data_list(StringDataValue), False),
        Field("server", StringDataValue, False),
        Field("remoteServer", StringDataValue, False),
        Field("caCerts", StringDataValue, False),
        Field("snapChannel", StringDataValue, False),
        Field("pypiURL", StringDataValue, False),
        Field("url", StringDataValue, False),
    ]

    def __init__(
        self,
        additionalPackages: Optional[List[str]],
        aptURL: Optional[str],
        suites: Optional[List[str]],
        server: Optional[str],
        remoteServer: Optional[str],
        caCerts: Optional[str],
        snapChannel: Optional[str],
        pypiURL: Optional[str],
        url: Optional[str],
    ):
        self.additionalPackages = additionalPackages
        self.aptURL = aptURL
        self.suites = suites
        self.server = server
        self.remoteServer = remoteServer
        self.caCerts = caCerts
        self.snapChannel = snapChannel
        self.pypiURL = pypiURL
        self.url = url


class OverrideSelector(DataObject):
    fields = [
        Field("series", StringDataValue, False),
        Field("cloud", StringDataValue, False),
        Field("variant", StringDataValue, False),
    ]

    def __init__(
        self,
        series: Optional[str],
        cloud: Optional[str],
        variant: Optional[str],
    ):
        self.series = series
        self.cloud = cloud
        self.variant = variant


class Override(DataObject):
    fields = [
        Field("selector", OverrideSelector, False),
        Field("affordances", Affordances, False),
        Field("obligations", Obligations, False),
        Field("directives", Directives, False),
    ]

    def __init__(
        self,
        selector: Optional[OverrideSelector],
        affordances: Optional[Affordances],
        obligations: Optional[Obligations],
        directives: Optional[Directives],
    ):
        self.selector = selector
        self.affordances = affordances
        self.obligations = obligations
        self.directives = directives


class Entitlement(DataObject):
    fields = [
        Field("entitled", BoolDataValue, False),
        Field("type", StringDataValue, False),
        Field("affordances", Affordances, False),
        Field("obligations", Obligations, False),
        Field("directives", Directives, False),
        Field("overrides", data_list(Override), False),
    ]

    def __init__(
        self,
        entitled: Optional[bool],
        type: Optional[str],
        affordances: Optional[Affordances],
        obligations: Optional[Obligations],
        directives: Optional[Directives],
        overrides: Optional[List[Override]],
    ):
        self.entitled = entitled
        self.type = type
        self.affordances = affordances
        self.obligations = obligations
        self.directives = directives
        self.overrides = overrides


class ContractInfo(DataObject):
    fields = [
        Field("name", StringDataValue, False),
        Field("id", StringDataValue, False),
        Field("createdAt", DatetimeDataValue, False),
        Field("createdBy", StringDataValue, False),
        Field("resourceEntitlements", data_list(Entitlement), False),
        Field("specificResourceEntitlements", data_list(Entitlement), False),
        Field("effectiveFrom", DatetimeDataValue, False),
        Field("effectiveTo", DatetimeDataValue, False),
        Field("products", data_list(StringDataValue), False),
        Field("origin", StringDataValue, False),
    ]

    def __init__(
        self,
        name: Optional[str],
        id: Optional[str],
        createdAt: Optional[str],
        createdBy: Optional[str],
        resourceEntitlements: Optional[List[Entitlement]],
        specificResourceEntitlements: Optional[List[Entitlement]],
        effectiveFrom: Optional[str],
        effectiveTo: Optional[str],
        products: Optional[List[str]],
        origin: Optional[str],
    ):
        self.name = name
        self.id = id
        self.createdAt = createdAt
        self.createdBy = createdBy
        self.resourceEntitlements = resourceEntitlements
        self.specificResourceEntitlements = specificResourceEntitlements
        self.effectiveFrom = effectiveFrom
        self.effectiveTo = effectiveTo
        self.products = products
        self.origin = origin


class MachineTokenInfo(DataObject):
    fields = [
        Field("machineId", StringDataValue, False),
        Field("accountInfo", AccountInfo, False),
        Field("contractInfo", ContractInfo, False),
        Field("expires", DatetimeDataValue, False),
    ]

    def __init__(
        self,
        machineId: Optional[str],
        accountInfo: Optional[AccountInfo],
        contractInfo: Optional[ContractInfo],
        expires: Optional[str],
    ):
        self.machineId = machineId
        self.accountInfo = accountInfo
        self.contractInfo = contractInfo
        self.expires = expires


class PublicMachineTokenData(DataObject):
    fields = [
        Field("activityInfo", ActivityInfo, False),
        Field("machineTokenInfo", MachineTokenInfo, False),
        Field("availableResources", data_list(AvailableResource), False),
    ]

    def __init__(
        self,
        activityInfo: Optional[ActivityInfo],
        machineTokenInfo: Optional[MachineTokenInfo],
        availableResources: Optional[List[AvailableResource]],
    ):
        self.activityInfo = activityInfo
        self.availableResources = availableResources
        self.machineTokenInfo = machineTokenInfo



import datetime
import logging

import mock
import pytest

from lib import timer
from lib.timer import MeteringTimedJob, TimedJob, run_jobs
from uaclient.exceptions import (
    InvalidFileEncodingError,
    InvalidFileFormatError,
)


class TestTimedJob:
    @pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
    @pytest.mark.parametrize("return_value", (True, False))
    def test_run_job_returns_true_on_successful_job_run(
        self, return_value, caplog_text, FakeConfig
    ):
        """Return True on successful job run."""
        config = FakeConfig()

        def success_job(cfg):
            assert cfg == config
            return return_value

        job = TimedJob("day_job", success_job, 14400)
        assert True is job.run(config)

        if return_value:
            # Job executed
            assert "Executed job: day_job" in caplog_text()
        else:
            # Job noops
            assert "Executed job: day_job" not in caplog_text()

    @pytest.mark.parametrize("caplog_text", [logging.WARNING], indirect=True)
    def test_run_job_returns_false_on_failed_job(
        self, FakeConfig, caplog_text
    ):
        """Return False on failed job run and warns in log."""
        cfg = FakeConfig()

        def failed_job(cfg):
            raise Exception("Something broke")

        job = TimedJob("day_job", failed_job, 14400)
        assert False is job.run(cfg)
        assert "Error executing job day_job: Something broke" in caplog_text()

    @pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
    @pytest.mark.parametrize("is_cfg_set", (False, True))
    def test_get_default_run_interval(
        self, is_cfg_set, caplog_text, FakeConfig
    ):
        """Use the default run interval when config is absent or invalid."""
        cfg = FakeConfig()
        if is_cfg_set:
            setattr(cfg, "day_job_timer", -3)
        job = TimedJob("day_job", lambda: None, 14400)

        assert 14400 == job.run_interval_seconds(cfg)
        if is_cfg_set:
            assert (
                "Invalid value for day_job interval found in config."
                in caplog_text()
            )

    def test_get_configured_run_interval(self, FakeConfig):
        """Use the configured run interval when not overriden."""
        cfg = FakeConfig()
        setattr(cfg, "day_job_timer", 28800)
        job = TimedJob("day_job", lambda: None, 14400)

        assert 28800 == job.run_interval_seconds(cfg)

    @pytest.mark.parametrize(
        "disabled_by_default,interval_value", ((False, 14400), (True, 0))
    )
    @pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
    def test_does_not_run_if_disabled(
        self, FakeConfig, caplog_text, disabled_by_default, interval_value
    ):
        cfg = FakeConfig()
        if not disabled_by_default:
            setattr(cfg, "day_job_timer", 0)

        m_disabled_job = mock.Mock()
        job = TimedJob("day_job", m_disabled_job, interval_value)
        assert False is job.run(cfg)
        assert 0 == m_disabled_job.call_count


class TestTimer:
    @pytest.mark.parametrize("has_next_run", (False, True))
    def test_run_jobs_persists_job_status_on_successful_run(
        self, has_next_run, FakeConfig
    ):
        """Successful job run results in updated job-status.json."""
        cfg = FakeConfig()
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        # we lose microseconds when deserializing
        now = now - datetime.timedelta(microseconds=now.microsecond)
        fake_file = mock.MagicMock()
        m_job_status = mock.MagicMock(next_run=None, last_run=None)

        if has_next_run:
            next_run = now - datetime.timedelta(seconds=1)
            m_job_status.next_run = next_run

        fake_file.read.return_value = mock.MagicMock(
            metering=m_job_status,
            update_messaging=None,
        )
        expected_next_run = now + datetime.timedelta(seconds=43200)

        m_job_func = mock.Mock()
        m_jobs = TimedJob("fake_job", m_job_func, 43200)

        with mock.patch("lib.timer.metering_job", m_jobs):
            with mock.patch.object(timer, "timer_jobs_state_file", fake_file):
                run_jobs(cfg, now)

        actual_job_state = fake_file.read().metering
        assert expected_next_run == actual_job_state.next_run
        assert [mock.call(cfg=cfg)] == m_job_func.call_args_list

    def test_run_job_ignores_late_next_run(self, FakeConfig):
        """Do not run if next_run points to future time."""
        cfg = FakeConfig()
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        # we lose microseconds when deserializing
        now = now - datetime.timedelta(microseconds=now.microsecond)

        fake_file = mock.MagicMock()
        m_job_status = mock.MagicMock(
            next_run=now + datetime.timedelta(seconds=14400), last_run=None
        )
        fake_file.read.return_value = mock.MagicMock(
            metering=m_job_status,
            update_messaging=None,
        )

        m_job_func = mock.Mock()
        m_job = TimedJob("metering", m_job_func, 43200)

        with mock.patch("lib.timer.metering_job", m_job):
            with mock.patch.object(timer, "timer_jobs_state_file", fake_file):
                run_jobs(cfg, now)

        actual_job_state = fake_file.read().metering
        assert actual_job_state.last_run is None
        assert 0 == m_job_func.call_count

    @pytest.mark.parametrize(
        "delete_side_effect", (None, OSError(), PermissionError())
    )
    @pytest.mark.parametrize(
        "read_side_effect",
        (
            InvalidFileFormatError(file_name="file", file_format="json"),
            InvalidFileEncodingError(file_name="file", file_encoding="utf-8"),
        ),
    )
    @pytest.mark.parametrize("caplog_text", [logging.WARNING], indirect=True)
    @mock.patch("lib.timer.AllTimerJobsState")
    @mock.patch("lib.timer.run_job")
    def test_run_jobs_tries_to_create_new_file_when_invalid(
        self,
        m_run_job,
        m_jobs_state,
        delete_side_effect,
        read_side_effect,
        caplog_text,
        FakeConfig,
    ):
        cfg = FakeConfig()
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        # we lose microseconds when deserializing
        now = now - datetime.timedelta(microseconds=now.microsecond)

        fake_file = mock.MagicMock()
        fake_file.read.side_effect = read_side_effect
        fake_file.delete.side_effect = delete_side_effect

        with mock.patch.object(timer, "timer_jobs_state_file", fake_file):
            run_jobs(cfg, now)

        assert [mock.call()] == fake_file.delete.call_args_list
        if delete_side_effect is None:
            assert [
                mock.call(m_jobs_state())
            ] == fake_file.write.call_args_list
            assert 2 == m_run_job.call_count
        else:
            assert [] == fake_file.write.call_args_list
            assert 0 == m_run_job.call_count
            log = caplog_text()
            assert (
                "Error trying to delete invalid jobs-status.json file" in log
            )


class TestMeteringTimedJob:
    @pytest.mark.parametrize(
        "activity_ping_interval_value,config_value,expected_value",
        ((None, 100, 100), (20, 80, 80), (1000, 80, 1000), (1000, 0, 0)),
    )
    @mock.patch("lib.timer.TimedJob.run_interval_seconds")
    def test_metering_run_interval_seconds(
        self,
        m_run_interval_seconds,
        activity_ping_interval_value,
        config_value,
        expected_value,
        fake_machine_token_file,
    ):
        m_run_interval_seconds.return_value = config_value
        m_cfg = mock.MagicMock()
        fake_machine_token_file.attached = True
        fake_machine_token_file.token = {
            "activityInfo": {
                "activityPingInterval": activity_ping_interval_value
            }
        }

        metering_job = MeteringTimedJob(
            job_func=mock.MagicMock(),
            default_interval_seconds=mock.MagicMock(),
        )
        assert expected_value == metering_job.run_interval_seconds(m_cfg)



import enum
import textwrap
from collections import defaultdict
from typing import Dict, List, Optional, Type

from uaclient import exceptions
from uaclient.config import UAConfig
from uaclient.entitlements import fips
from uaclient.entitlements.anbox import AnboxEntitlement
from uaclient.entitlements.base import UAEntitlement
from uaclient.entitlements.cc import CommonCriteriaEntitlement
from uaclient.entitlements.cis import CISEntitlement
from uaclient.entitlements.entitlement_status import ApplicabilityStatus
from uaclient.entitlements.esm import ESMAppsEntitlement, ESMInfraEntitlement
from uaclient.entitlements.esm_legacy import (
    ESMAppsLegacyEntitlement,
    ESMInfraLegacyEntitlement,
)
from uaclient.entitlements.landscape import LandscapeEntitlement
from uaclient.entitlements.livepatch import LivepatchEntitlement
from uaclient.entitlements.realtime import RealtimeKernelEntitlement
from uaclient.entitlements.repo import RepoEntitlement
from uaclient.entitlements.ros import ROSEntitlement, ROSUpdatesEntitlement
from uaclient.exceptions import EntitlementNotFoundError

ENTITLEMENT_CLASSES = [
    AnboxEntitlement,
    CommonCriteriaEntitlement,
    CISEntitlement,
    ESMAppsEntitlement,
    ESMAppsLegacyEntitlement,
    ESMInfraEntitlement,
    ESMInfraLegacyEntitlement,
    fips.FIPSEntitlement,
    fips.FIPSUpdatesEntitlement,
    fips.FIPSPreviewEntitlement,
    LandscapeEntitlement,
    LivepatchEntitlement,
    RealtimeKernelEntitlement,
    ROSEntitlement,
    ROSUpdatesEntitlement,
]  # type: List[Type[UAEntitlement]]


def entitlement_factory(
    cfg: UAConfig,
    name: str,
    variant: str = "",
    purge: bool = False,
    access_only: bool = False,
    extra_args: Optional[List[str]] = None,
) -> UAEntitlement:
    """Returns a UAEntitlement object based on the provided name.

    :param cfg: UAConfig instance
    :param name: The name of the entitlement to return
    :param variant: The variant name to be used
    :param purge: If purge operation is enabled
    :param access_only: If entitlement should be set with access only
    :param extra_args: Extra parameters to create the entitlement

    :raise EntitlementNotFoundError: If no entitlement with the given name is
      found, or if the entitlement exists but no variant with the specified
      name is found.
    """

    for entitlement in ENTITLEMENT_CLASSES:
        ent = entitlement(
            cfg=cfg,
            access_only=access_only,
            called_name=name,
            purge=purge,
            extra_args=extra_args,
        )
        if name in ent.valid_names:
            if not variant:
                return ent
            elif variant in ent.variants:
                return ent.variants[variant](
                    cfg=cfg,
                    called_name=name,
                    purge=purge,
                    extra_args=extra_args,
                )
            else:
                raise EntitlementNotFoundError(entitlement_name=variant)
    raise EntitlementNotFoundError(entitlement_name=name)


def valid_services(cfg: UAConfig, all_names: bool = False) -> List[str]:
    """Return a list of valid services.

    :param cfg: UAConfig instance
    :param all_names: if we should return all the names for a service instead
        of just the presentation_name
    """
    entitlements = ENTITLEMENT_CLASSES
    if all_names:
        names = []
        for entitlement_cls in entitlements:
            names.extend(entitlement_cls(cfg=cfg).valid_names)

        return sorted(names)

    return sorted(
        [
            entitlement_cls(cfg=cfg).presentation_name
            for entitlement_cls in entitlements
        ]
    )


def order_entitlements_for_enabling(
    cfg: UAConfig, ents: List[str]
) -> List[str]:
    """
    A function to sort entitlments for enabling that preserves invalid names
    """
    valid_ents_ordered = entitlements_enable_order(cfg)

    def sort_order_with_nonexistent_last(ent):
        try:
            return valid_ents_ordered.index(ent)
        except ValueError:
            return len(valid_ents_ordered)

    return sorted(ents, key=lambda ent: sort_order_with_nonexistent_last(ent))


@enum.unique
class SortOrder(enum.Enum):
    REQUIRED_SERVICES = object()
    DEPENDENT_SERVICES = object()


def entitlements_disable_order(cfg: UAConfig) -> List[str]:
    """
    Return the entitlements disable order based on dependent services logic.
    """
    return _sort_entitlements(cfg=cfg, sort_order=SortOrder.DEPENDENT_SERVICES)


def entitlements_enable_order(cfg: UAConfig) -> List[str]:
    """
    Return the entitlements enable order based on required services logic.
    """
    return _sort_entitlements(cfg=cfg, sort_order=SortOrder.REQUIRED_SERVICES)


def _sort_entitlements_visit(
    cfg: UAConfig,
    ent_cls: Type[UAEntitlement],
    sort_order: SortOrder,
    visited: Dict[str, bool],
    order: List[str],
):
    if ent_cls.name in visited:
        return

    ent = ent_cls(cfg)

    if sort_order == SortOrder.REQUIRED_SERVICES:
        cls_list = [e.entitlement for e in ent.required_services]
    else:
        cls_list = list(ent.dependent_services)

    for cls_dependency in cls_list:
        if ent_cls.name not in visited:
            _sort_entitlements_visit(
                cfg=cfg,
                ent_cls=cls_dependency,
                sort_order=sort_order,
                visited=visited,
                order=order,
            )

    order.append(str(ent_cls.name))
    visited[str(ent_cls.name)] = True


def _sort_entitlements(cfg: UAConfig, sort_order: SortOrder) -> List[str]:
    order = []  # type: List[str]
    visited = {}  # type: Dict[str, bool]

    for ent_cls in ENTITLEMENT_CLASSES:
        _sort_entitlements_visit(
            cfg=cfg,
            ent_cls=ent_cls,
            sort_order=sort_order,
            visited=visited,
            order=order,
        )

    return order


def get_valid_entitlement_names(names: List[str], cfg: UAConfig):
    """Return a list of valid entitlement names.

    :param names: List of entitlements to validate
    :return: a tuple of List containing the valid and invalid entitlements
    """
    entitlements_found = []

    for ent_name in names:
        if ent_name in valid_services(cfg=cfg, all_names=True):
            entitlements_found.append(ent_name)

    entitlements_not_found = sorted(set(names) - set(entitlements_found))

    return entitlements_found, entitlements_not_found


def create_enable_entitlements_not_found_error(
    entitlements_not_found, cfg: UAConfig
) -> exceptions.UbuntuProError:
    """
    Constructs the MESSAGE_INVALID_SERVICE_OP_FAILURE message
    based on the attempted services and valid services.
    """
    valid_services_names = valid_services(cfg=cfg)
    valid_names = ", ".join(valid_services_names)
    service_msg = "\n".join(
        textwrap.wrap(
            "Try " + valid_names + ".",
            width=80,
            break_long_words=False,
            break_on_hyphens=False,
        )
    )
    return exceptions.InvalidServiceOpError(
        operation="enable",
        invalid_service=", ".join(entitlements_not_found),
        service_msg=service_msg,
    )


def check_entitlement_apt_directives_are_unique(
    cfg: UAConfig,
) -> bool:
    entitlement_directives = defaultdict(list)

    for ent_name in valid_services(cfg):
        ent = entitlement_factory(cfg, ent_name)

        if not isinstance(ent, RepoEntitlement):
            continue

        applicability_status, _ = ent.applicability_status()

        if applicability_status == ApplicabilityStatus.APPLICABLE:
            apt_url = ent.apt_url
            apt_suites = ent.apt_suites or ()

            for suite in apt_suites:
                entitlement_directive = ent.repo_policy_check_tmpl.format(
                    apt_url, suite
                )
                entitlement_directives[entitlement_directive].append(
                    {
                        "from": ent_name,
                        "apt_url": apt_url,
                        "suite": suite,
                    }
                )

        for def_path, ent_directive in entitlement_directives.items():
            if len(ent_directive) > 1:
                ent_apt_url = ent_directive[0]["apt_url"]
                ent_suite = ent_directive[0]["suite"]

                raise exceptions.EntitlementsAPTDirectivesAreNotUnique(
                    url=cfg.contract_url,
                    names=", ".join(
                        sorted(str(ent["from"]) for ent in ent_directive)
                    ),
                    apt_url=ent_apt_url,
                    suite=ent_suite,
                )

    return True


def get_title(cfg: UAConfig, ent_name: str, variant=""):
    try:
        return entitlement_factory(cfg, ent_name, variant=variant).title
    except exceptions.UbuntuProError:
        return ent_name



from subprocess import TimeoutExpired

import mock
import pytest

from uaclient import exceptions
from uaclient.daemon import start, stop

M_PATH = "uaclient.daemon."


@mock.patch(M_PATH + "system.subp")
class TestStart:
    def test_start_success(self, m_subp):
        start()
        assert [
            mock.call(
                ["systemctl", "start", "ubuntu-advantage.service"], timeout=2.0
            )
        ] == m_subp.call_args_list

    @pytest.mark.parametrize(
        "err",
        (
            exceptions.ProcessExecutionError("cmd"),
            TimeoutExpired("cmd", 2.0),
        ),
    )
    @mock.patch(M_PATH + "LOG.warning")
    def test_start_warning(self, m_log_warning, m_subp, err):
        m_subp.side_effect = err
        start()
        assert [
            mock.call(
                ["systemctl", "start", "ubuntu-advantage.service"], timeout=2.0
            )
        ] == m_subp.call_args_list
        assert [mock.call(err, exc_info=err)] == m_log_warning.call_args_list


@mock.patch(M_PATH + "system.subp")
class TestStop:
    def test_stop_success(self, m_subp):
        stop()
        assert [
            mock.call(
                ["systemctl", "stop", "ubuntu-advantage.service"], timeout=2.0
            )
        ] == m_subp.call_args_list

    @pytest.mark.parametrize(
        "err",
        (
            exceptions.ProcessExecutionError("cmd"),
            TimeoutExpired("cmd", 2.0),
        ),
    )
    @mock.patch(M_PATH + "LOG.warning")
    def test_stop_warning(self, m_log_warning, m_subp, err):
        m_subp.side_effect = err
        stop()
        assert [
            mock.call(
                ["systemctl", "stop", "ubuntu-advantage.service"], timeout=2.0
            )
        ] == m_subp.call_args_list
        assert [mock.call(err, exc_info=err)] == m_log_warning.call_args_list



from typing import Any, Dict, Optional, Tuple

from uaclient import api, contract, event_logger, messages, system
from uaclient.entitlements.entitlement_status import (
    CanEnableFailure,
    CanEnableFailureReason,
)
from uaclient.entitlements.repo import RepoEntitlement
from uaclient.files.state_files import (
    AnboxCloudData,
    anbox_cloud_credentials_file,
)
from uaclient.types import MessagingOperationsDict

event = event_logger.get_event_logger()


class AnboxEntitlement(RepoEntitlement):
    name = "anbox-cloud"
    title = messages.ANBOX_TITLE
    description = messages.ANBOX_DESCRIPTION
    help_doc_url = messages.urls.ANBOX_HOME_PAGE
    help_text = messages.ANBOX_HELP_TEXT
    repo_key_file = "ubuntu-pro-anbox-cloud.gpg"
    repo_url_tmpl = "{}"
    affordance_check_series = True
    supports_access_only = True
    origin = "Anbox"

    @property
    def messaging(self) -> MessagingOperationsDict:
        if not self.access_only:
            return {"post_enable": [messages.ANBOX_RUN_INIT_CMD]}
        else:
            return {}

    def can_enable(self) -> Tuple[bool, Optional[CanEnableFailure]]:
        ret, reason = super().can_enable()

        if not ret:
            return ret, reason

        if system.is_container() and not self.access_only:
            return (
                False,
                CanEnableFailure(
                    CanEnableFailureReason.ONLY_ACCESS_ONLY_SUPPORTED,
                    messages.ANBOX_FAIL_TO_ENABLE_ON_CONTAINER,
                ),
            )

        return True, None

    def _perform_enable(self, progress: api.ProgressWrapper) -> bool:
        ret = super()._perform_enable(progress)

        if not ret:
            return ret

        directives = self.entitlement_cfg.get("entitlement", {}).get(
            "directives", {}
        )
        machine_token = self.machine_token_file.machine_token["machineToken"]
        client = contract.UAContractClient(self.cfg)
        anbox_images_machine_access = client.get_resource_machine_access(
            machine_token, "anbox-images"
        )

        anbox_cloud_data = AnboxCloudData(
            anbox_images_url=anbox_images_machine_access.get("entitlement", {})
            .get("directives", {})
            .get("url", ""),
            anbox_images_resource_token=anbox_images_machine_access.get(
                "resourceToken", ""
            ),
            anbox_cloud_apt_url=directives.get("aptURL", ""),
            anbox_cloud_apt_token=directives.get("aptKey", ""),
        )

        anbox_cloud_credentials_file.write(anbox_cloud_data)

        return True

    def _perform_disable(self, progress: api.ProgressWrapper):
        super()._perform_disable(progress)
        anbox_cloud_credentials_file.delete()
        return True

    # TODO: remove this function
    # This is just a placeholder until we can deliver the Anbox
    # resourceToken from the contracts without relying on the
    # enableByDefault obligation
    def _should_enable_by_default(
        self, obligations: Dict[str, Any], resourceToken: Optional[str]
    ) -> bool:
        return False



import logging
from typing import List, Optional

from uaclient import entitlements, lock, messages, status, util
from uaclient.api import AbstractProgress, ProgressWrapper, exceptions
from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.api.u.pro.status.enabled_services.v1 import _enabled_services
from uaclient.api.u.pro.status.is_attached.v1 import _is_attached
from uaclient.config import UAConfig
from uaclient.data_types import (
    BoolDataValue,
    DataObject,
    Field,
    StringDataValue,
    data_list,
)

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))


class DisableOptions(DataObject):
    fields = [
        Field("service", StringDataValue, doc="Pro service to disable"),
        Field(
            "purge",
            BoolDataValue,
            False,
            doc=(
                "Also remove all packages that were installed from this"
                " service. Only supported by some services. (default: false)"
            ),
        ),
    ]

    def __init__(self, *, service: str, purge: bool = False):
        self.service = service
        self.purge = purge


class DisableResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "disabled",
            data_list(StringDataValue),
            doc="List of services disabled",
        ),
    ]

    def __init__(self, *, disabled: List[str]):
        self.disabled = disabled


def _enabled_services_names(cfg: UAConfig) -> List[str]:
    return [s.name for s in _enabled_services(cfg).enabled_services]


def disable(
    options: DisableOptions, progress_object: Optional[AbstractProgress] = None
) -> DisableResult:
    return _disable(options, UAConfig(), progress_object=progress_object)


def _disable(
    options: DisableOptions,
    cfg: UAConfig,
    progress_object: Optional[AbstractProgress] = None,
) -> DisableResult:
    """
    Disable a Pro service. This will automatically disable any services that
    depend on the target service.
    """
    progress = ProgressWrapper(progress_object)

    if not util.we_are_currently_root():
        raise exceptions.NonRootUserError()

    if not _is_attached(cfg).is_attached:
        raise exceptions.UnattachedError()

    entitlement = entitlements.entitlement_factory(
        cfg=cfg,
        name=options.service,
        purge=options.purge,
    )

    enabled_services_before = _enabled_services_names(cfg)

    # Do this after getting the class so that the factory can raise an
    # exception for invalid service names
    if options.service not in enabled_services_before:
        # nothing to do
        return DisableResult(
            disabled=[],
        )

    variant = entitlement.enabled_variant
    if variant is not None:
        entitlement = variant

    progress.total_steps = entitlement.calculate_total_disable_steps()

    success = False
    fail_reason = None

    try:
        with lock.RetryLock(
            lock_holder="u.pro.services.disable.v1",
        ):
            success, fail_reason = entitlement.disable(progress)
    except Exception as e:
        lock.clear_lock_file_if_present()
        raise e

    if not success:
        if fail_reason is not None and fail_reason.message is not None:
            reason = fail_reason.message
        else:
            reason = messages.GENERIC_UNKNOWN_ISSUE
        raise exceptions.EntitlementNotDisabledError(
            service=options.service, reason=reason
        )

    enabled_services_after = _enabled_services_names(cfg)

    status.status(cfg=cfg)  # Update the status cache
    progress.finish()

    return DisableResult(
        disabled=sorted(
            list(
                set(enabled_services_before).difference(
                    set(enabled_services_after)
                )
            )
        ),
    )


endpoint = APIEndpoint(
    version="v1",
    name="DisableService",
    fn=_disable,
    options_cls=DisableOptions,
    supports_progress=True,
)

_doc = {
    "introduced_in": "32",
    "requires_network": True,
    "example_python": """
from uaclient.api.u.pro.services.disable.v1 import disable, DisableOptions
result = disable(DisableOptions(service="usg"))
""",
    "result_class": DisableResult,
    "exceptions": [
        (exceptions.NonRootUserError, "When called as non-root user"),
        (
            exceptions.UnattachedError,
            (
                "When called on a machine that is not attached to a Pro"
                " subscription"
            ),
        ),
        (
            exceptions.EntitlementNotFoundError,
            "When the service argument is not a valid Pro service name",
        ),
        (
            exceptions.LockHeldError,
            "When another Ubuntu Pro related operation is in progress",
        ),
        (
            exceptions.EntitlementNotDisabledError,
            "When the service fails to disable",
        ),
    ],
    "example_cli": "pro api u.pro.services.disable.v1 --args service=usg",
    "example_json": """
{
    "disabled": [
        "usg"
    ]
}
""",
}
